"""OpenAPI spec renderer: render_paths."""

import textwrap

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


def test_render_paths(testrenderer, oas_fragment):
    """Usual paths definition is rendered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences/{evidenceId}:
                  summary: Ignored
                  description: Ignored
                  servers:
                    url: https://example.com
                  parameters:
                    - name: evidenceId
                      in: path
                      required: true
                      description: A unique evidence identifier to query.
                      schema:
                        type: string
                  get:
                    summary: Retrieve an evidence by ID.
                    description: More verbose description...
                    parameters:
                      - name: details
                        in: query
                        description: If true, information w/ details is returned.
                        schema:
                          type: boolean
                    responses:
                      '200':
                        description: An evidence.
                      '404':
                        description: An evidence not found.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences/{evidenceId}

           **Retrieve an evidence by ID.**

           More verbose description...

           :param evidenceId:
              A unique evidence identifier to query.
           :paramtype evidenceId: string, required
           :queryparam details:
              If true, information w/ details is returned.
           :queryparamtype details: boolean
           :statuscode 200:
              An evidence.
           :statuscode 404:
              An evidence not found.
        """
    )


def test_render_paths_minimal(testrenderer, oas_fragment):
    """Minimal paths definition is rendered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences:
                  get:
                    responses:
                      '200':
                        description: A list of evidences.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences

           :statuscode 200:
              A list of evidences.
        """
    )


def test_render_paths_multiple(testrenderer, oas_fragment):
    """Paths definition with multiple paths is rendered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences/{evidenceId}:
                  get:
                    summary: Retrieve an evidence by ID.
                    description: More verbose description...
                    parameters:
                      - name: evidenceId
                        in: path
                        required: true
                        description: A unique evidence identifier to query.
                        schema:
                          type: string
                      - name: details
                        in: query
                        description: If true, information w/ details is returned.
                        schema:
                          type: boolean
                    responses:
                      '200':
                        description: An evidence.
                      '404':
                        description: An evidence not found.
                /evidences:
                  post:
                    summary: Create an evidence.
                    responses:
                      '201':
                        description: An evidence created.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences/{evidenceId}

           **Retrieve an evidence by ID.**

           More verbose description...

           :param evidenceId:
              A unique evidence identifier to query.
           :paramtype evidenceId: string, required
           :queryparam details:
              If true, information w/ details is returned.
           :queryparamtype details: boolean
           :statuscode 200:
              An evidence.
           :statuscode 404:
              An evidence not found.

        .. http:post:: /evidences

           **Create an evidence.**

           :statuscode 201:
              An evidence created.
        """
    )


def test_render_paths_parameters_common(testrenderer, oas_fragment):
    """Paths definition with common parameters is rendered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences/{evidenceId}:
                  get:
                    summary: Retrieve an evidence by ID.
                    parameters:
                      - name: details
                        in: query
                        description: If true, information w/ details is returned.
                        schema:
                          type: boolean
                    responses:
                      '200':
                        description: An evidence.
                      '404':
                        description: An evidence not found.
                  put:
                    summary: Update an evidence by ID.
                    responses:
                      '200':
                        description: An evidence.
                      '404':
                        description: An evidence not found.
                  parameters:
                    - name: evidenceId
                      in: path
                      required: true
                      description: A unique evidence identifier to query.
                      schema:
                        type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences/{evidenceId}

           **Retrieve an evidence by ID.**

           :param evidenceId:
              A unique evidence identifier to query.
           :paramtype evidenceId: string, required
           :queryparam details:
              If true, information w/ details is returned.
           :queryparamtype details: boolean
           :statuscode 200:
              An evidence.
           :statuscode 404:
              An evidence not found.

        .. http:put:: /evidences/{evidenceId}

           **Update an evidence by ID.**

           :param evidenceId:
              A unique evidence identifier to query.
           :paramtype evidenceId: string, required
           :statuscode 200:
              An evidence.
           :statuscode 404:
              An evidence not found.
        """
    )


def test_render_paths_parameters_common_prepend(testrenderer, oas_fragment):
    """Paths definition with common parameters is rendered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences/{evidenceId}/{evidenceSection}:
                  get:
                    summary: Retrieve an evidence by ID.
                    parameters:
                      - name: evidenceSection
                        in: path
                        description: Query a section with a given name.
                        schema:
                          type: string
                    responses:
                      '200':
                        description: An evidence.
                      '404':
                        description: An evidence not found.
                  parameters:
                    - name: evidenceId
                      in: path
                      required: true
                      description: A unique evidence identifier to query.
                      schema:
                        type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences/{evidenceId}/{evidenceSection}

           **Retrieve an evidence by ID.**

           :param evidenceId:
              A unique evidence identifier to query.
           :paramtype evidenceId: string, required
           :param evidenceSection:
              Query a section with a given name.
           :paramtype evidenceSection: string
           :statuscode 200:
              An evidence.
           :statuscode 404:
              An evidence not found.
        """
    )


def test_render_paths_parameters_common_overwritten(testrenderer, oas_fragment):
    """Paths definition with common parameters is rendered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences/{evidenceId}:
                  get:
                    summary: Retrieve an evidence by ID.
                    parameters:
                      - name: evidenceId
                        in: path
                        description: Overwritten description.
                        schema:
                          type: string
                    responses:
                      '200':
                        description: An evidence.
                  parameters:
                    - name: evidenceId
                      in: path
                      required: true
                      description: A unique evidence identifier to query.
                      schema:
                        type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences/{evidenceId}

           **Retrieve an evidence by ID.**

           :param evidenceId:
              Overwritten description.
           :paramtype evidenceId: string
           :statuscode 200:
              An evidence.
        """
    )


def test_render_paths_methods_order(testrenderer, oas_fragment):
    """Paths definition is rendered with HTTP methods ordered."""

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences:
                  post:
                    responses:
                      '201':
                        description: An evidence created.
                  options:
                    responses:
                      '200':
                        description: CORS preflight request.
                  get:
                    responses:
                      '200':
                        description: A list of evidences.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           :statuscode 201:
              An evidence created.

        .. http:options:: /evidences

           :statuscode 200:
              CORS preflight request.

        .. http:get:: /evidences

           :statuscode 200:
              A list of evidences.
        """
    )


def test_render_paths_methods_order_custom(fakestate, oas_fragment):
    """Paths definition is rendered with HTTP methods ordered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"http-methods-order": ["delete", "options", "get", "post"]}
    )

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences:
                  post:
                    responses:
                      '201':
                        description: An evidence created.
                  options:
                    responses:
                      '200':
                        description: CORS preflight request.
                  get:
                    responses:
                      '200':
                        description: A list of evidences.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:options:: /evidences

           :statuscode 200:
              CORS preflight request.

        .. http:get:: /evidences

           :statuscode 200:
              A list of evidences.

        .. http:post:: /evidences

           :statuscode 201:
              An evidence created.
        """
    )


def test_render_paths_methods_order_insensitive(fakestate, oas_fragment):
    """Paths definition is rendered with HTTP methods ordered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"http-methods-order": ["gEt", "post"]}
    )

    markup = textify(
        testrenderer.render_paths(
            oas_fragment(
                """
                /evidences:
                  post:
                    responses:
                      '201':
                        description: An evidence created.
                  get:
                    responses:
                      '200':
                        description: A list of evidences.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences

           :statuscode 200:
              A list of evidences.

        .. http:post:: /evidences

           :statuscode 201:
              An evidence created.
        """
    )



import pathlib
import os

import pytest


@pytest.fixture(scope="function")
def rendered():
    return pathlib.Path(os.path.dirname(__file__), "rendered")


@pytest.mark.regenerate_rendered_specs
def test_generate(testrenderer, testspec, rendered):
    testspec_name, testspec = testspec
    rendered_markup = "\n".join(testrenderer.render_restructuredtext_markup(testspec))

    rendered.joinpath(os.path.dirname(testspec_name)).mkdir(parents=True, exist_ok=True)
    rendered.joinpath(testspec_name + ".rst").write_text(rendered_markup)


def test_render(testrenderer, testspec, rendered):
    testspec_name, testspec = testspec
    rendered_markup = "\n".join(testrenderer.render_restructuredtext_markup(testspec))

    assert rendered_markup == rendered.joinpath(testspec_name + ".rst").read_text()



""".convert_responses() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_responses(oas_fragment):
    def _wrapper(responses, produces):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"]["responses"] = responses
        oas2["paths"]["/test"]["get"]["produces"] = produces

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"]["responses"]

    return _wrapper


def test_minimal(convert_responses, oas_fragment):
    converted = convert_responses(
        oas_fragment(
            """
            '200':
              description: a response description
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        '200':
          description: a response description
        """
    )


def test_complete(convert_responses, oas_fragment):
    converted = convert_responses(
        oas_fragment(
            """
            '200':
              description: a response description
              examples:
                application/json:
                  something: important
              headers:
                X-Test:
                  description: Is it a test?
                  type: string
              schema:
                items:
                  format: int32
                  type: integer
                type: array
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        '200':
          description: a response description
          content:
            application/json:
              example:
                something: important
              schema:
                items:
                  format: int32
                  type: integer
                type: array
          headers:
            X-Test:
              description: Is it a test?
              schema:
                type: string
        """
    )


def test_multiple(convert_responses, oas_fragment):
    converted = convert_responses(
        oas_fragment(
            """
            '200':
              description: OK
              schema:
                items:
                  format: int32
                  type: integer
                type: array
            '400':
              description: Bad Request
              headers:
                X-Test:
                  description: Is it a test?
                  type: string
            default:
              description: Internal Server Error
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        '200':
          content:
            application/json:
              schema:
                items:
                  format: int32
                  type: integer
                type: array
          description: OK
        '400':
          description: Bad Request
          headers:
            X-Test:
              description: Is it a test?
              schema:
                type: string
        default:
          description: Internal Server Error
        """
    )


def test_vendor_extensions(convert_responses, oas_fragment):
    converted = convert_responses(
        oas_fragment(
            """
            '200':
              description: a response description
            x-vendor-ext: vendor-ext
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        '200':
          description: a response description
        x-vendor-ext: vendor-ext
        """
    )



"""Here lies OpenAPI renderers."""

from . import abc
from ._httpdomain_old import HttpdomainOldRenderer
from ._httpdomain import HttpdomainRenderer
from ._model import ModelRenderer
from ._toc import TocRenderer


__all__ = [
    "abc",
    "HttpdomainOldRenderer",
    "HttpdomainRenderer",
    "ModelRenderer",
    "TocRenderer",
]



"""OpenAPI spec renderer: render_parameters."""

import itertools

import textwrap
import pytest

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


def test_render_parameters_no_items(testrenderer, oas_fragment):
    """No parameter definitions are rendered."""

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                []
                """
            )
        )
    )
    assert markup == ""


def test_render_parameters_one_item(testrenderer, oas_fragment):
    """One usual parameter definition is rendered."""

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                - name: evidenceId
                  in: path
                  required: true
                  description: A unique evidence identifier to query.
                  schema:
                    type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        """.rstrip()
    )


def test_render_parameters_many_items(testrenderer, oas_fragment):
    """Many parameter definitions are rendered."""

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                - name: evidenceId
                  in: path
                  required: true
                  description: A unique evidence identifier to query.
                  schema:
                    type: string
                - name: details
                  in: query
                  description: If true, information w/ details is returned.
                  schema:
                    type: boolean
                - name: Api-Version
                  in: header
                  default: '1'
                  description: API version to use for the request.
                  schema:
                    type: integer
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader Api-Version:
           API version to use for the request.
        :reqheadertype Api-Version: integer
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        :queryparam details:
           If true, information w/ details is returned.
        :queryparamtype details: boolean
        """.rstrip()
    )


@pytest.mark.parametrize("permutation_seq", itertools.permutations(range(3)))
def test_render_parameters_many_items_ordered(
    testrenderer, oas_fragment, permutation_seq
):
    """Many parameter definitions are rendered and properly ordered."""

    parameters = oas_fragment(
        """
        - name: evidenceId
          in: path
          required: true
          description: A unique evidence identifier to query.
          schema:
            type: string
        - name: details
          in: query
          description: If true, information w/ details is returned.
          schema:
            type: boolean
        - name: Api-Version
          in: header
          required: false
          default: '1'
          description: API version to use for the request.
          schema:
            type: integer
        """
    )

    markup = textify(
        testrenderer.render_parameters(
            # Since the test receives a permutation sequence as input,
            # we need to ensure that parameters are shuffled according
            # to that sequence, because this is the essence of the test.
            [parameters[seq] for seq in permutation_seq]
        )
    )

    assert markup == textwrap.dedent(
        """\
        :reqheader Api-Version:
           API version to use for the request.
        :reqheadertype Api-Version: integer
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        :queryparam details:
           If true, information w/ details is returned.
        :queryparamtype details: boolean
        """.rstrip()
    )


def test_render_parameters_many_items_stable_order(testrenderer, oas_fragment):
    """Many parameter definitions are rendered w/ preserved order."""

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                - name: kind
                  in: path
                  required: true
                  description: An evidence kind.
                  schema:
                    type: string
                - name: Api-Version
                  in: header
                  default: '1'
                  description: API version to use for the request.
                  schema:
                    type: integer
                - name: details
                  in: query
                  description: If true, information w/ details is returned.
                  schema:
                    type: boolean
                - name: evidenceId
                  in: path
                  required: true
                  description: A unique evidence identifier to query.
                  schema:
                    type: string
                - name: related
                  in: query
                  description: If true, links to related evidences are returned.
                  schema:
                    type: boolean
                - name: Accept
                  in: header
                  default: application/json
                  description: A desired Content-Type of HTTP response.
                  schema:
                    type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader Api-Version:
           API version to use for the request.
        :reqheadertype Api-Version: integer
        :reqheader Accept:
           A desired Content-Type of HTTP response.
        :reqheadertype Accept: string
        :param kind:
           An evidence kind.
        :paramtype kind: string, required
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        :queryparam details:
           If true, information w/ details is returned.
        :queryparamtype details: boolean
        :queryparam related:
           If true, links to related evidences are returned.
        :queryparamtype related: boolean
        """.rstrip()
    )


def test_render_parameters_custom_order(fakestate, oas_fragment):
    """Many parameter definitions are rendered w/ preserved order."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"request-parameters-order": ["query", "path", "header"]}
    )

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                - name: kind
                  in: path
                  required: true
                  description: An evidence kind.
                  schema:
                    type: string
                - name: Api-Version
                  in: header
                  default: '1'
                  description: API version to use for the request.
                  schema:
                    type: integer
                - name: details
                  in: query
                  description: If true, information w/ details is returned.
                  schema:
                    type: boolean
                - name: evidenceId
                  in: path
                  required: true
                  description: A unique evidence identifier to query.
                  schema:
                    type: string
                - name: related
                  in: query
                  description: If true, links to related evidences are returned.
                  schema:
                    type: boolean
                - name: Accept
                  in: header
                  default: application/json
                  description: A desired Content-Type of HTTP response.
                  schema:
                    type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam details:
           If true, information w/ details is returned.
        :queryparamtype details: boolean
        :queryparam related:
           If true, links to related evidences are returned.
        :queryparamtype related: boolean
        :param kind:
           An evidence kind.
        :paramtype kind: string, required
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        :reqheader Api-Version:
           API version to use for the request.
        :reqheadertype Api-Version: integer
        :reqheader Accept:
           A desired Content-Type of HTTP response.
        :reqheadertype Accept: string
        """.rstrip()
    )


def test_render_parameters_custom_order_partial(fakestate, oas_fragment):
    """Many parameter definitions are rendered w/ preserved order."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"request-parameters-order": ["query", "path"]}
    )

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                - name: kind
                  in: path
                  required: true
                  description: An evidence kind.
                  schema:
                    type: string
                - name: Api-Version
                  in: header
                  default: '1'
                  description: API version to use for the request.
                  schema:
                    type: integer
                - name: details
                  in: query
                  description: If true, information w/ details is returned.
                  schema:
                    type: boolean
                - name: evidenceId
                  in: path
                  required: true
                  description: A unique evidence identifier to query.
                  schema:
                    type: string
                - name: related
                  in: query
                  description: If true, links to related evidences are returned.
                  schema:
                    type: boolean
                - name: Accept
                  in: header
                  default: application/json
                  description: A desired Content-Type of HTTP response.
                  schema:
                    type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam details:
           If true, information w/ details is returned.
        :queryparamtype details: boolean
        :queryparam related:
           If true, links to related evidences are returned.
        :queryparamtype related: boolean
        :param kind:
           An evidence kind.
        :paramtype kind: string, required
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        :reqheader Api-Version:
           API version to use for the request.
        :reqheadertype Api-Version: integer
        :reqheader Accept:
           A desired Content-Type of HTTP response.
        :reqheadertype Accept: string
        """.rstrip()
    )


def test_render_parameters_case_insensitive(fakestate, oas_fragment):
    """Many parameter definitions are rendered w/ preserved order."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"request-parameters-order": ["QUERY", "pAth", "Header"]}
    )

    markup = textify(
        testrenderer.render_parameters(
            oas_fragment(
                """
                - name: kind
                  in: PATH
                  required: true
                  description: An evidence kind.
                  schema:
                    type: string
                - name: Api-Version
                  in: header
                  default: '1'
                  description: API version to use for the request.
                  schema:
                    type: integer
                - name: details
                  in: query
                  description: If true, information w/ details is returned.
                  schema:
                    type: boolean
                - name: evidenceId
                  in: Path
                  required: true
                  description: A unique evidence identifier to query.
                  schema:
                    type: string
                - name: related
                  in: qUery
                  description: If true, links to related evidences are returned.
                  schema:
                    type: boolean
                - name: Accept
                  in: headeR
                  default: application/json
                  description: A desired Content-Type of HTTP response.
                  schema:
                    type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam details:
           If true, information w/ details is returned.
        :queryparamtype details: boolean
        :queryparam related:
           If true, links to related evidences are returned.
        :queryparamtype related: boolean
        :param kind:
           An evidence kind.
        :paramtype kind: string, required
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        :reqheader Api-Version:
           API version to use for the request.
        :reqheadertype Api-Version: integer
        :reqheader Accept:
           A desired Content-Type of HTTP response.
        :reqheadertype Accept: string
        """.rstrip()
    )




import re
from . import abc
from .. import utils
from docutils.parsers.rst import directives


class TocRenderer(abc.RestructuredTextRenderer):

    option_spec = {
        # nb columns
        "nb_columns": directives.positive_int,
    }

    def __init__(self, state, options):
        self._state = state
        self._options = options
        if 'nb_columns' not in self._options:
            self._options["nb_columns"] = 2

    def render_restructuredtext_markup(self, spec):

        utils.normalize_spec(spec, **self._options)

        yield ""
        yield ".. hlist::"
        yield "    :columns: {}".format(self._options["nb_columns"])
        yield ""

        for path in spec["paths"].keys():
            cpath = re.sub(r"[{}]", "", re.sub(r"[<>:/]", "-", path))
            for verb, ope in spec["paths"][path].items():
                yield "    - `{} <#{}>`_".format(
                    ope.get("operationId", verb + " " + path),
                    verb.lower() + "-" + cpath
                )
        yield ""



from . import abc
from .. import utils

import re
import hashlib
import json
import textwrap

import jsonschema
from jsonschema import validate
from docutils.parsers.rst import directives

from sphinx.locale import get_translation
_ = get_translation('openapi')


def _get_description(obj, convert):
    D = convert(obj.get('description', '')).strip()
    if 'default' in obj:
        if D and D[-1] != '.':
            D += '.'
        if D:
            D += ' '
        D += 'Default: ``' + json.dumps(obj['default']) + "``."
    return D


def _get_contraints(obj):
    c = []
    if 'minItems' in obj:
        c.append(_('minItems is {}').format(obj['minItems']))
    if 'maxItems' in obj:
        c.append(_('maxItems is {}').format(obj['maxItems']))
    if 'minLength' in obj:
        c.append(_('minLength is {}').format(obj['minLength']))
    if 'maxLength' in obj:
        c.append(_('maxLength is {}').format(obj['maxLength']))
    if 'minimum' in obj:
        c.append(_('minimum is {}').format(obj['minimum']))
    if 'maximum' in obj:
        c.append(_('maximum is {}').format(obj['maximum']))
    if 'uniqueItems' in obj:
        c.append(_("items must be unique"))
    if "pattern" in obj:
        c.append(_("pattern ``{}``").format(obj["pattern"]))
    if 'enum' in obj:
        c.append(_('possible values are {}').format(
                 ', '.join(
                    [
                        '``{}``'.format(x) for x in obj['enum']
                    ]
                 )))
    if 'readOnly' in obj:
        c.append(_("read only"))
    if 'writeOnly' in obj:
        c.append(_("write only"))
    s = '; '.join(c)
    return s


def _add_constraints(obj, D, C):
    if C:
        if _('Constraints') not in D:
            C = _("Constraints: {}").format(C)
            if D and D[-1] != '.':
                D += '.'
        else:
            if C and D and D[-1] != ';':
                D += ';'
        if D and C and C[0] != '\n':
            D += ' '
        D += C
    else:
        if D and D[-1] != '.':
            D += '.'
    if 'deprecated' in obj:
        D += "\n\n**{}**".format(_("DEPRECATED"))
    return D


def _get_multi_type(schema, entities):
    T = []
    duplicate = set()
    if 'oneOf' in schema:
        k = 'oneOf'
    elif 'allOf' in schema:
        k = 'allOf'
    else:
        k = 'anyOf'
    for t in schema[k]:
        type = t.get('type', 'object')
        if '$entity_ref' in t and type == 'object':
            T.append(ref2link(entities, t['$entity_ref']))
        else:
            if type == 'string' and 'enum' in t:
                type = 'enumerate'
                vals = ', '.join(
                    [
                        '``{}``'.format(x) for x in t['enum']
                    ]
                )
                if vals:
                    type += ' (' + vals + ')'
            if type not in duplicate:
                T.append(type)
                duplicate.add(type)
    return T


def _process_one(prefix, schema, mandatory, entities, convert):
    if 'oneOf' in schema:
        type = 'oneOf'
    elif 'allOf' in schema:
        type = 'allOf'
    elif 'anyOf' in schema:
        type = 'anyOf'
    else:
        type = schema.get('type', 'object')
    if '$entity_ref' in schema and type == 'object' and prefix:
        # does not apply to first level types (prefix empty)
        T = _('Object of type {}').format(ref2link(entities, schema['$entity_ref']))
        D = _get_description(schema, convert)
        C = _get_contraints(schema)
        D = _add_constraints(schema, D, C)
        ret = ['.'.join(prefix), T, D, mandatory]
        yield ret
    elif type == 'array':
        ref = schema['items'].get('$entity_ref', None)
        type_items = schema['items'].get('type', None)
        if ref:
            D = _get_description(schema, convert)
            C = _get_contraints(schema)
            D = _add_constraints(schema, D, C)
            if type_items not in ['object', 'array']:
                # Since only object and array are described on their own
                # Add here the constraints of the referenced type
                D = _get_description(schema['items'], convert)
                C = _get_contraints(schema['items'])
                D = _add_constraints(schema['items'], D, C)
                C = _get_contraints(schema)
                D = _add_constraints(schema, D, C)
                yield [
                    '.'.join(prefix),
                    _('Array of {}').format(type_items),
                    D,
                    mandatory
                ]
            else:
                yield [
                    '.'.join(prefix),
                    _('Array of {}').format(ref2link(entities, ref)),
                    D,
                    mandatory
                ]
        elif type_items == 'object':
            T = _("Array")
            D = _get_description(schema, convert)
            C = _get_contraints(schema)
            D = _add_constraints(schema, D, C)
            yield ['.'.join(prefix), T, D, mandatory]
            if prefix:
                prefix[-1] += '[]'
            else:
                prefix = ['[]']
            for x in _process_one(prefix, schema['items'], False, entities, convert):
                yield x
        else:
            # Support array of simple types (string, etc.)
            D = _get_description(schema, convert)
            C = _get_contraints(schema)
            for x in _process_one(prefix, schema['items'], False, entities, convert):
                # Add C to x[2] now and not before (to avoid double "Constraints:")
                if D and x[2]:
                    DD = D + ' ' + x[2]
                else:
                    DD = D + x[2]
                DD = _add_constraints(schema, DD, C)
                yield [x[0], _('Array of {}').format(x[1]), DD, mandatory]
    elif type == 'object':
        required = schema.get('required', [])
        for prop_name, prop in schema.get('properties', {}).items():
            for x in _process_one(
                    prefix+[prop_name],
                    prop,
                    prop_name in required,
                    entities,
                    convert):
                yield x
        if 'additionalProperties' in schema:
            D = _('Additional properties')
            if schema['additionalProperties'] is True:
                T = ''
            elif schema['additionalProperties'] is False:
                return
            else:
                if 'oneOf' in schema['additionalProperties']:
                    T = _("One of {}").format(
                        ", ".join(_get_multi_type(schema['additionalProperties'], entities)))
                elif 'allOf' in schema['additionalProperties']:
                    T = _("All of {}").format(
                        ", ".join(_get_multi_type(schema['additionalProperties'], entities)))
                elif 'anyOf' in schema['additionalProperties']:
                    T = _("Any of {}").format(
                        ", ".join(_get_multi_type(schema['additionalProperties'], entities)))
                else:
                    T = schema['additionalProperties'].get('type', 'object')
            yield ['.'.join(prefix+['*']), T, D, '']
    elif 'oneOf' in schema:
        # One of the subtype, must be basic types or ref
        D = _get_description(schema, convert)
        C = _get_contraints(schema)
        D = _add_constraints(schema, D, C)
        T = _get_multi_type(schema, entities)
        T = _("One of {}").format(", ".join(T))
        yield ['.'.join(prefix), T, D, mandatory]
    elif 'allOf' in schema:
        # All of the subtype, must be basic types or ref
        D = _get_description(schema, convert)
        C = _get_contraints(schema)
        D = _add_constraints(schema, D, C)
        T = _get_multi_type(schema, entities)
        T = _("All of {}").format(", ".join(T))
        yield ['.'.join(prefix), T, D, mandatory]
    elif 'anyOf' in schema:
        # Any of the subtype, must be basic types or ref
        D = _get_description(schema, convert)
        C = _get_contraints(schema)
        D = _add_constraints(schema, D, C)
        T = _get_multi_type(schema, entities)
        T = _("Any of {}").format(", ".join(T))
        yield ['.'.join(prefix), T, D, mandatory]
    elif type in ['string', 'integer', 'number', 'boolean']:
        T = type
        if schema.get('format', ''):
            T += '/' + schema.get('format', '')
        D = _get_description(schema, convert)
        C = _get_contraints(schema)
        D = _add_constraints(schema, D, C)
        yield ['.'.join(prefix), T, D, mandatory]


def _build(name, schema, entities, convert, options):
    if 'type' not in schema:
        schema['type'] = 'object'
    if schema.get('type', '') not in ['object', 'array']:
        return ''

    yield ''
    yield '.. _'+entities('/components/schemas/'+name)+':'
    yield ''
    yield name
    yield options['header'] * len(name)
    yield ''
    D = _get_description(schema, convert)
    if D:
        yield D
        yield ''
    yield '.. list-table:: ' + name
    yield '    :header-rows: 1'
    yield '    :widths: 25 25 45 15'
    yield '    :class: longtable'
    yield ''
    yield '    * - ' + _('Attribute')
    yield '      - ' + _('Type')
    yield '      - ' + _('Description')
    yield '      - ' + _('Required')

    for item in _process_one([], schema, False, entities, convert):
        if str(item[0]):
            yield '    * - ``' + str(item[0]) + '``'
        else:
            yield '    * - ' + _('N/A')
        yield '      - ' + textwrap.indent(str(item[1]), '        ').lstrip()
        yield '      - ' + textwrap.indent(str(item[2]), '        ').lstrip()
        yield '      - ' + _('Yes') if item[3] else '      -'

    if 'example' in schema or 'examples' in schema:
        N = 1
        for ex in [schema.get('example', None)] + schema.get('examples', []):
            if ex is None:
                continue
            yield ''
            yield _('Example #{}:').format(N)
            N += 1
            # validate the example against this schema
            try:
                validate(instance=ex, schema=schema)
                yield ''
                yield '.. code-block:: json'
                yield ''
                for line in json.dumps(ex, indent=2).splitlines():
                    yield '    ' + line
            except jsonschema.ValidationError:
                yield ''
                yield '**{}**'.format(_('Invalid example'))


def ref2link(entities, ref):
    if ref in ['object', 'string']:
        return ref
    name = ref.split('/')[-1]
    if ref[0] == '#':
        ref = ref[1:]
    if callable(entities):
        ref = entities(ref)
        return ':ref:`{name} <{ref}>`'.format(**locals())
    else:
        return '{name}'.format(**locals())


def _entities(spec, ref):
    m = hashlib.md5()
    m.update(spec.get('info', {}).get('title', '').encode('utf-8'))
    m.update(spec.get('info', {}).get('version', '0.0').encode('utf-8'))
    key = m.hexdigest()
    # for unit tests
    if key == '30565a8911a6bb487e3745c0ea3c8224':
        key = ''
    if '#' in ref:
        ref = ref.split('#')[1]
    return key + ref


class ModelRenderer(abc.RestructuredTextRenderer):

    option_spec = {
        # prefix (components/schemas)
        "prefix": str,
        # header marker (')
        "header": directives.single_char_or_unicode,
        # Markup format to render OpenAPI descriptions.
        "format": str,
        # A list of entities to be rendered. Must be whitespace delimited.
        "entities": lambda s: s.split(),
        # Regular expression patterns to include/exclude entities to/from
        # rendering. The patterns must be whitespace delimited.
        "include": lambda s: s.split(),
        "exclude": lambda s: s.split(),
    }

    def __init__(self, state, options):
        self._state = state
        self._options = options
        if 'header' not in self._options:
            self._options["header"] = "'"
        if 'prefix' not in self._options:
            self._options["prefix"] = "/components/schemas"

    def render_restructuredtext_markup(self, spec):

        utils.normalize_spec(spec, **self._options)

        convert = utils.get_text_converter(self._options)

        schemas = spec
        for p in filter(None, self._options["prefix"].split('/')):
            schemas = schemas.get(p, {})

        # Entities list to be processed
        entities = []

        # If 'entities' are passed we've got to ensure they exist within an OpenAPI
        # spec; otherwise raise error and ask user to fix that.
        if 'entities' in self._options:
            if not set(self._options['entities']).issubset(schemas.keys()):
                raise ValueError(
                    'One or more entities are not defined in the spec: %s.' % (
                        ', '.join(set(self._options['entities']) - set(schemas.keys())),
                    )
                )
            entities = self._options['entities']

        # Check against regular expressions to be included
        if 'include' in self._options:
            for i in self._options['include']:
                ir = re.compile(i)
                for entity in schemas.keys():
                    if ir.match(entity):
                        entities.append(entity)

        # If no include nor entities option, then take full entity
        if 'include' not in self._options and 'entities' not in self._options:
            entities = schemas.keys()

        # Remove entities matching regexp
        if 'exclude' in self._options:
            tmp_entities = []
            for e in self._options['exclude']:
                er = re.compile(e)
                for entity in entities:
                    if not er.match(entity):
                        tmp_entities.append(entity)
            entities = tmp_entities

        def __entities(x):
            return _entities(spec, x)

        for name in entities:
            schema = schemas[name]
            for line in _build(name, schema, __entities, convert, self._options):
                yield line.rstrip()
            yield ''



"""OpenAPI spec renderer: render_request_body_example."""

import textwrap

import pytest
import responses

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


@pytest.mark.parametrize(
    ["content"],
    [
        pytest.param(
            """
            content:
              application/json:
                examples:
                  test:
                    value:
                      foo: bar
                      baz: 42
            """,
            id="examples",
        ),
        pytest.param(
            """
            content:
              application/json:
                example:
                  foo: bar
                  baz: 42
            """,
            id="example",
        ),
        pytest.param(
            """
            content:
              application/json:
                schema:
                  example:
                    foo: bar
                    baz: 42
            """,
            id="schema/example",
        ),
        pytest.param(
            """
            content:
              application/json:
                examples:
                  test:
                    value: |
                      {
                        "foo": "bar",
                        "baz": 42
                      }
            """,
            id="examples::str",
        ),
        pytest.param(
            """
            content:
              application/json:
                example: |
                  {
                    "foo": "bar",
                    "baz": 42
                  }
            """,
            id="example::str",
        ),
        pytest.param(
            """
            content:
              application/json:
                schema:
                  example: |
                    {
                      "foo": "bar",
                      "baz": 42
                    }
            """,
            id="schema/example::str",
        ),
        pytest.param(
            """
            content:
              application/json:
                schema:
                  example:
                    foobar: bazinga
                example:
                  foo: bar
                  baz: 42
            """,
            id="example-beats-schema/example",
        ),
        pytest.param(
            """
            content:
              application/json:
                schema:
                  example:
                    foobar: bazinga
                examples:
                  test:
                    value:
                      foo: bar
                      baz: 42
            """,
            id="examples-beats-schema/example",
        ),
    ],
)
def test_render_request_body_example(testrenderer, content, oas_fragment):
    """Request body is rendered."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(content), "/evidences/{evidenceId}", "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: application/json

           {
             "foo": "bar",
             "baz": 42
           }
        """.rstrip()
    )


def test_render_request_body_example_1st_from_examples(testrenderer, oas_fragment):
    """Request body's first example is rendered."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    examples:
                      foo:
                        value:
                          foo: bar
                          baz: 42
                      bar:
                        value:
                          foobar: bazinga
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: application/json

           {
             "foo": "bar",
             "baz": 42
           }
        """.rstrip()
    )


def test_render_request_body_example_1st_from_media_type(testrenderer, oas_fragment):
    """Request body's example from first media type is rendered."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  text/plain:
                    example: |
                      foo = "bar"
                      baz = 42
                  application/json:
                    schema:
                      type: object
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["example_preference_key"],
    [pytest.param("request-example-preference"), pytest.param("example-preference")],
)
def test_render_request_body_example_preference(
    fakestate, oas_fragment, example_preference_key
):
    """Request body's example from preferred media type is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {example_preference_key: ["text/plain"]}
    )

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    example:
                      foo: bar
                      baz: 42
                  text/plain:
                    example: |
                      foo = "bar"
                      baz = 42
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["example_preference_key"],
    [pytest.param("request-example-preference"), pytest.param("example-preference")],
)
def test_render_request_body_example_preference_complex(
    fakestate, oas_fragment, example_preference_key
):
    """Request body's example from preferred media type is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {example_preference_key: ["application/json", "text/plain"]}
    )

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  text/csv:
                    example: |
                      foo,baz
                      bar,42
                  text/plain:
                    example: |
                      foo = "bar"
                      baz = 42
                  application/json:
                    schema:
                      type: object
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


def test_render_request_body_example_preference_priority(fakestate, oas_fragment):
    """Request body's example from preferred media type is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate,
        {
            "example-preference": ["application/json"],
            "request-example-preference": ["text/plain"],
        },
    )

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    example:
                      foo: bar
                      baz: 42
                  text/plain:
                    example: |
                      foo = "bar"
                      baz = 42
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


@responses.activate
def test_render_request_body_example_external(testrenderer, oas_fragment):
    """Request body's example can be retrieved from external location."""

    responses.add(
        responses.GET,
        "https://example.com/json/examples/test.json",
        json={"foo": "bar", "baz": 42},
        status=200,
    )

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    examples:
                      test:
                        externalValue: https://example.com/json/examples/test.json
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: application/json

           {"foo": "bar", "baz": 42}
        """.rstrip()
    )


@responses.activate
def test_render_request_body_example_external_errored_next_example(
    testrenderer, oas_fragment, caplog
):
    """Request body's example fallbacks on next when external cannot be retrieved."""

    responses.add(
        responses.GET, "https://example.com/json/examples/test.json", status=404,
    )

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    examples:
                      test:
                        externalValue: https://example.com/json/examples/test.json
                      fallback:
                        value: '{"spam": 42}'
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: application/json

           {"spam": 42}
        """.rstrip()
    )


@responses.activate
def test_render_request_body_example_external_errored_next_media_type(
    testrenderer, oas_fragment, caplog,
):
    """Request body's example fallbacks on next when external cannot be retrieved."""

    responses.add(
        responses.GET, "https://example.com/json/examples/test.json", status=404,
    )

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    examples:
                      test:
                        externalValue: https://example.com/json/examples/test.json
                  text/csv:
                    example: spam,42
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: text/csv

           spam,42
        """.rstrip()
    )


def test_render_request_body_example_content_type(testrenderer, oas_fragment):
    """Request body's example can render something other than application/json."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  text/csv:
                    example: |
                      foo,baz
                      bar,42
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           POST /evidences/{evidenceId} HTTP/1.1
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )


def test_render_request_body_example_noop(testrenderer, oas_fragment):
    """Request body's example is not rendered if there's nothing to render."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  application/json:
                    schema:
                      type: object
                """
            ),
            "/evidences/{evidenceId}",
            "POST",
        )
    )

    assert markup == ""


@pytest.mark.parametrize(
    ["http_method"], [pytest.param("POST"), pytest.param("PUT"), pytest.param("PATCH")]
)
def test_render_request_body_example_http_method(
    testrenderer, oas_fragment, http_method,
):
    """Request body's example shows proper HTTP method."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  text/csv:
                    example: |
                      foo,baz
                      bar,42
                """
            ),
            "/evidences/{evidenceId}",
            http_method,
        )
    )

    assert markup == textwrap.dedent(
        f"""\
        .. sourcecode:: http

           {http_method} /evidences/{{evidenceId}} HTTP/1.1
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["http_endpoint"],
    [pytest.param("/evidences/{evidenceId}"), pytest.param("/heroes/{heroId}")],
)
def test_render_request_body_example_http_endpoint(
    testrenderer, oas_fragment, http_endpoint,
):
    """Request body's example shows proper HTTP method."""

    markup = textify(
        testrenderer.render_request_body_example(
            oas_fragment(
                """
                content:
                  text/csv:
                    example: |
                      foo,baz
                      bar,42
                """
            ),
            http_endpoint,
            "POST",
        )
    )

    assert markup == textwrap.dedent(
        f"""\
        .. sourcecode:: http

           POST {http_endpoint} HTTP/1.1
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )



"""Abstract Base Classes (ABCs) for OpenAPI renderers."""

import abc

from docutils import nodes
from docutils.statemachine import ViewList
from sphinx.util.nodes import nested_parse_with_titles


class Renderer(metaclass=abc.ABCMeta):
    """Base class for OpenAPI renderers."""

    def __init__(self, state, options):
        self._state = state
        self._options = options

    @property
    @abc.abstractmethod
    def option_spec(self):
        """Renderer options and their converting functions."""

    @abc.abstractmethod
    def render(self, spec):
        """Render a given OpenAPI spec."""


class RestructuredTextRenderer(Renderer):
    """Base class for reStructuredText OpenAPI renderers.

    Docutils DOM manipulation is quite a tricky task that requires passing
    dozen arguments around. Because of that a lot of Sphinx extensions instead
    of constructing DOM nodes directly produce and parse reStructuredText.
    This Sphinx extension is not an exception, and that's why this class
    exists. It's a convenient extension of :class:`Renderer` that converts
    produced markup text into docutils DOM elements.
    """

    def render(self, spec):
        viewlist = ViewList()
        for line in self.render_restructuredtext_markup(spec):
            viewlist.append(line, "<openapi>")

        node = nodes.section()
        node.document = self._state.document
        nested_parse_with_titles(self._state, viewlist, node)
        return node.children



""".convert() test suite."""

import sphinxcontrib.openapi._lib2to3 as lib2to3


def test_minimal(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_complete(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            tags:
              - tag_a
            externalDocs: https://docs.example.com/
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        tags:
          - tag_a
        externalDocs: https://docs.example.com/
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_servers_complete(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            host: example.com
            basePath: /v1
            schemes:
              - https
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        servers:
          - url: https://example.com/v1
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_servers_host_only(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            host: example.com
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        servers:
          - url: example.com
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_servers_basepath_only(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            basePath: /v1
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        servers:
          - url: /v1
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_servers_schemes_multiple(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            host: example.com
            schemes:
              - http
              - https
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        servers:
          - url: http://example.com
          - url: https://example.com
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_servers_schemes_from_operation(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            host: example.com
            schemes:
              - http
            paths:
              /test:
                get:
                  schemes:
                    - ws
                  responses:
                    '200':
                      description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        servers:
          - url: http://example.com
          - url: ws://example.com
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        """
    )


def test_consumes(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            consumes:
              - application/json
            paths:
              /test:
                post:
                  parameters:
                    - in: query
                      name: marker
                      type: string
                    - in: body
                      name: payload
                      schema:
                        type: string
                  responses:
                    '201':
                      description: a response description
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        paths:
          /test:
            post:
              parameters:
                - in: query
                  name: marker
                  schema:
                    type: string
              requestBody:
                content:
                  application/json:
                    schema:
                      type: string
              responses:
                '201':
                  description: a response description
        """
    )


def test_consumes_operation_override(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            consumes:
              - application/xml
            paths:
              /test:
                post:
                  consumes:
                    - application/json
                  parameters:
                    - in: query
                      name: marker
                      type: string
                    - in: body
                      name: payload
                      schema:
                        type: string
                  responses:
                    '201':
                      description: a response description
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        paths:
          /test:
            post:
              parameters:
                - in: query
                  name: marker
                  schema:
                    type: string
              requestBody:
                content:
                  application/json:
                    schema:
                      type: string
              responses:
                '201':
                  description: a response description
        """
    )


def test_produces(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            produces:
              - application/json
            paths:
              /test:
                get:
                  responses:
                    '200':
                      schema:
                        items:
                          format: int32
                          type: integer
                        type: array
                      description: a response description
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        paths:
          /test:
            get:
              responses:
                '200':
                  content:
                    application/json:
                      schema:
                        items:
                          format: int32
                          type: integer
                        type: array
                  description: a response description
        """
    )


def test_produces_operation_override(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            produces:
              - application/xml
            paths:
              /test:
                get:
                  produces:
                    - application/json
                  responses:
                    '200':
                      schema:
                        items:
                          format: int32
                          type: integer
                        type: array
                      description: a response description
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        paths:
          /test:
            get:
              responses:
                '200':
                  content:
                    application/json:
                      schema:
                        items:
                          format: int32
                          type: integer
                        type: array
                  description: a response description
        """
    )


def test_vendor_extensions(oas_fragment):
    converted = lib2to3.convert(
        oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: 1.0
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            x-vendor-ext: vendor-ext
            """
        )
    )
    assert converted == oas_fragment(
        """
        openapi: 3.0.3
        info:
          title: An example spec
          version: 1.0
        paths:
          /test:
            get:
              responses:
                '200':
                  description: a response description
        x-vendor-ext: vendor-ext
        """
    )



import textwrap
import collections

import yaml
from sphinxcontrib.openapi import renderers


class TestOpenApi3HttpDomain(object):

    def test_basic(self):
        renderer = renderers.ModelRenderer(None, {})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {},
            'components': {
                'schemas': {
                    'Resource': {
                        'type': 'object',
                        'required': ['kind'],
                        'properties': collections.OrderedDict([
                            ('kind', {
                                'description': 'Kind',
                                'type': 'string',
                                'readOnly': True,
                            }),
                            ('instance', {
                                '$ref': '#/components/schemas/Instance',
                            }),
                        ]),
                    },
                    'Instance': {
                        'type': 'object',
                        'properties': collections.OrderedDict([
                            ('instance', {
                                'description': 'Instance',
                                'type': 'string',
                                'enum': ['A', 'B'],
                                'writeOnly': True,
                                'deprecated': True,
                            }),
                            ('instanceType', {
                                '$ref': '#/components/schemas/InstanceType',
                            }),
                            ('dep', {
                                'description': 'Deprecation',
                                'type': 'string',
                                'deprecated': True,
                            }),
                        ]),
                    },
                    'InstanceList': {
                        'type': 'array',
                        'items': {
                            '$ref':  '#/components/schemas/Instance'
                        }
                    },
                    'InstanceType': {
                        'type': 'string',
                        'enum': ['T1', 'T2'],
                    },
                },
            },
        }))

        assert text == textwrap.dedent("""
        .. _/components/schemas/Resource:

        Resource
        ''''''''

        .. list-table:: Resource
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``kind``
              - string
              - Kind. Constraints: read only
              - Yes
            * - ``instance``
              - Object of type :ref:`Instance </components/schemas/Instance>`
              -
              -


        .. _/components/schemas/Instance:

        Instance
        ''''''''

        .. list-table:: Instance
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``instance``
              - string
              - Instance. Constraints: possible values are ``A``, ``B``; write only

                **DEPRECATED**
              -
            * - ``instanceType``
              - string
              - Constraints: possible values are ``T1``, ``T2``
              -
            * - ``dep``
              - string
              - Deprecation.

                **DEPRECATED**
              -


        .. _/components/schemas/InstanceList:

        InstanceList
        ''''''''''''

        .. list-table:: InstanceList
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - N/A
              - Array of :ref:`Instance </components/schemas/Instance>`
              -
              -

        """)

    def test_options(self):
        renderer = renderers.ModelRenderer(None, {"header": "?", "prefix": "definitions"})
        text = '\n'.join(renderer.render_restructuredtext_markup({
            'openapi': '3.0.0',
            'paths': {},
            'definitions': {
                'Resource': {
                    'type': 'object',
                    'required': ['kind'],
                    'properties': collections.OrderedDict([
                        ('kind', {
                            'description': 'Kind',
                            'type': 'string',
                        }),
                    ]),
                },
            },
        }))

        assert text == textwrap.dedent("""
        .. _/components/schemas/Resource:

        Resource
        ????????

        .. list-table:: Resource
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``kind``
              - string
              - Kind.
              - Yes
        """)

    def test_types(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              properties:
                field1:
                  type: integer
                  format: int32
                  description: Signed 32 bits
                field2:
                  type: number
                  format: float
                  description: Float
                field3:
                  type: string
                  format: byte
                  description: base64 encoded characters
                field4:
                  type: boolean
                  default: false
                field5:
                  type: array
                  items:
                    type: string
                    maxLength: 255
                  minItems: 1
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - integer/int32
              - Signed 32 bits.
              -
            * - ``field2``
              - number/float
              - Float.
              -
            * - ``field3``
              - string/byte
              - base64 encoded characters.
              -
            * - ``field4``
              - boolean
              - Default: ``false``.
              -
            * - ``field5``
              - Array of string
              - Constraints: maxLength is 255; minItems is 1
              -
        """)

    def test_array(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: array
              items:
                type: object
                required:
                  - field1
                properties:
                  field1:
                    type: string
            Test2:
              type: object
              properties:
                table:
                  type: array
                  items:
                    type: object
                    required:
                      - field2
                    properties:
                      field2:
                        type: string
                        pattern: "a-zA-Z0-9"
                      field3:
                        type: array
                        items:
                          $ref: '#/components/schemas/Test1'
                        minItems: 1
                      field4:
                        type: array
                        items:
                          type: string
                        minItems: 1
                        maxItems: 10
                        uniqueItems: True
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - N/A
              - Array
              -
              -
            * - ``[].field1``
              - string
              -
              - Yes


        .. _/components/schemas/Test2:

        Test2
        '''''

        .. list-table:: Test2
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``table``
              - Array
              -
              -
            * - ``table[].field2``
              - string
              - Constraints: pattern ``a-zA-Z0-9``
              - Yes
            * - ``table[].field3``
              - Array of :ref:`Test1 </components/schemas/Test1>`
              - Constraints: minItems is 1
              -
            * - ``table[].field4``
              - Array of string
              - Constraints: minItems is 1; maxItems is 10; items must be unique
              -
        """)

    def test_markdown(self):
        renderer = renderers.ModelRenderer(None, {'format': 'markdown'})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              description: This is a __bold__ description
              type: object
              properties:
                field1:
                  type: integer
                  format: int32
                  description: Signed _32_ bits
                  default: 5
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        This is a **bold** description

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - integer/int32
              - Signed *32* bits. Default: ``5``.
              -
        """)

    def test_example(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              properties:
                field1:
                  type: integer
                  format: int32
              example:
                field1: 12
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - integer/int32
              -
              -

        Example #1:

        .. code-block:: json

            {
              "field1": 12
            }
        """)

    def test_examples(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              properties:
                field1:
                  type: integer
                  format: int32
              examples:
                - field1: 12
                - field1: -2
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - integer/int32
              -
              -

        Example #1:

        .. code-block:: json

            {
              "field1": 12
            }

        Example #2:

        .. code-block:: json

            {
              "field1": -2
            }
        """)

    def test_bad_example(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              properties:
                field1:
                  type: integer
                  format: int32
              example:
                field1: true
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - integer/int32
              -
              -

        Example #1:

        **Invalid example**
        """)

    def test_oneof(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              additionalProperties:
                oneOf:
                - type: string
                - type: integer
                - type: number
                - type: boolean
                - $ref: '#/components/schemas/Test4'
                - $ref: '#/components/schemas/Test5'
            Test2:
              type: object
              properties:
                field2:
                  oneOf:
                  - type: string
                  - type: integer
                  - type: number
                  - type: boolean
            Test3:
              oneOf:
                - $ref: '#/components/schemas/Test4'
                - type: object
                  required:
                    - field4
            Test4:
              type: object
              properties:
                field4:
                  type: string
              additionalProperties: false
            Test5:
              type: string
              enum: [A, B]
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``*``
              - One of string, integer, number, boolean, """ +
                                       """:ref:`Test4 </components/schemas/Test4>`, enumerate (``A``, ``B``)
              - Additional properties
              -


        .. _/components/schemas/Test2:

        Test2
        '''''

        .. list-table:: Test2
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field2``
              - One of string, integer, number, boolean
              -
              -


        .. _/components/schemas/Test3:

        Test3
        '''''

        .. list-table:: Test3
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - N/A
              - One of :ref:`Test4 </components/schemas/Test4>`, object
              -
              -


        .. _/components/schemas/Test4:

        Test4
        '''''

        .. list-table:: Test4
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field4``
              - string
              -
              -

        """)

    def test_allof(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              additionalProperties:
                allOf:
                - type: string
                - type: integer
                - type: number
                - type: boolean
                - $ref: '#/components/schemas/Test4'
            Test2:
              type: object
              properties:
                field2:
                  allOf:
                  - type: string
                  - type: integer
                  - type: number
                  - type: boolean
            Test3:
              allOf:
                - $ref: '#/components/schemas/Test4'
                - type: object
                  required:
                    - field4
            Test4:
              type: object
              properties:
                field4:
                  type: string
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``*``
              - All of string, integer, number, boolean, :ref:`Test4 </components/schemas/Test4>`
              - Additional properties
              -


        .. _/components/schemas/Test2:

        Test2
        '''''

        .. list-table:: Test2
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field2``
              - All of string, integer, number, boolean
              -
              -


        .. _/components/schemas/Test3:

        Test3
        '''''

        .. list-table:: Test3
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - N/A
              - All of :ref:`Test4 </components/schemas/Test4>`, object
              -
              -


        .. _/components/schemas/Test4:

        Test4
        '''''

        .. list-table:: Test4
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field4``
              - string
              -
              -
        """)

    def test_anyof(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              additionalProperties:
                anyOf:
                - type: string
                - type: integer
                - type: number
                - type: boolean
                - $ref: '#/components/schemas/Test4'
            Test2:
              type: object
              properties:
                field2:
                  anyOf:
                  - type: string
                  - type: integer
                  - type: number
                  - type: boolean
            Test3:
              anyOf:
                - $ref: '#/components/schemas/Test4'
                - type: object
                  required:
                    - field4
            Test4:
              type: object
              properties:
                field4:
                  type: string
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``*``
              - Any of string, integer, number, boolean, :ref:`Test4 </components/schemas/Test4>`
              - Additional properties
              -


        .. _/components/schemas/Test2:

        Test2
        '''''

        .. list-table:: Test2
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field2``
              - Any of string, integer, number, boolean
              -
              -


        .. _/components/schemas/Test3:

        Test3
        '''''

        .. list-table:: Test3
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - N/A
              - Any of :ref:`Test4 </components/schemas/Test4>`, object
              -
              -


        .. _/components/schemas/Test4:

        Test4
        '''''

        .. list-table:: Test4
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field4``
              - string
              -
              -
        """)

    def test_others_additional(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              required:
                - field1
              properties:
                field1:
                  type: string
                  example: F1
                others:
                  type: object
                  additionalProperties: true
              additionalProperties: false
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - string
              -
              - Yes
            * - ``others.*``
              -
              - Additional properties
              -
        """)

    def test_filtering(self):
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            A:
              properties:
                a:
                  type: string
            B:
              properties:
                b:
                  type: string
            AB:
              properties:
                ab:
                  type: string
        """))

        renderer = renderers.ModelRenderer(None, {"include": ["A.*"]})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/A:

        A
        '

        .. list-table:: A
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``a``
              - string
              -
              -


        .. _/components/schemas/AB:

        AB
        ''

        .. list-table:: AB
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``ab``
              - string
              -
              -
        """)

        renderer = renderers.ModelRenderer(None, {"include": ["A.*"], "exclude": [".*B"]})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/A:

        A
        '

        .. list-table:: A
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``a``
              - string
              -
              -
        """)

        renderer = renderers.ModelRenderer(None, {"entities": ["AB", "B"], "exclude": ["AB"]})
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/B:

        B
        '

        .. list-table:: B
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``b``
              - string
              -
              -
        """)

    def test_array_enum(self):
        renderer = renderers.ModelRenderer(None, {})
        spec = yaml.safe_load(textwrap.dedent("""
        ---
        openapi: 3.0.0
        paths: {}
        components:
          schemas:
            Test1:
              type: object
              properties:
                field1:
                  type: array
                  items:
                    $ref: '#/components/schemas/Type1'
                  minItems: 1
                field2:
                  type: array
                  items:
                    type: string
                    enum: [A, B]
                  minItems: 1
            Type1:
              type: string
              enum: [C, D]
        """))
        text = '\n'.join(renderer.render_restructuredtext_markup(spec))
        assert text == textwrap.dedent("""
        .. _/components/schemas/Test1:

        Test1
        '''''

        .. list-table:: Test1
            :header-rows: 1
            :widths: 25 25 45 15
            :class: longtable

            * - Attribute
              - Type
              - Description
              - Required
            * - ``field1``
              - Array of string
              - Constraints: possible values are ``C``, ``D``; minItems is 1
              -
            * - ``field2``
              - Array of string
              - Constraints: possible values are ``A``, ``B``; minItems is 1
              -

        """)



""".convert_paths() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_paths(oas_fragment):
    def _wrapper(paths):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"] = paths

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]

    return _wrapper


def test_minimal(convert_paths, oas_fragment):
    converted = convert_paths(
        oas_fragment(
            """
            /test:
              get:
                responses:
                  '200':
                    description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        /test:
          get:
            responses:
              '200':
                description: a response description
        """
    )


def test_complete(convert_paths, oas_fragment):
    pass
    converted = convert_paths(
        oas_fragment(
            """
            /{username}:
              parameters:
                - in: path
                  name: username
                  required: true
                  type: string
              get:
                tags:
                  - tag_a
                  - tag_b
                summary: an operation summary
                description: an operation description
                externalDocs: https://docs.example.com/
                operationId: myOperation
                produces:
                  - application/json
                parameters:
                  - in: header
                    name: token
                    type: string
                  - in: query
                    name: id
                    type: string
                responses:
                  '200':
                    schema:
                      items:
                        format: int32
                        type: integer
                      type: array
                    description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        /{username}:
          parameters:
            - in: path
              name: username
              schema:
                type: string
              required: true
          get:
            tags:
              - tag_a
              - tag_b
            summary: an operation summary
            description: an operation description
            externalDocs: https://docs.example.com/
            operationId: myOperation
            parameters:
              - in: header
                name: token
                schema:
                  type: string
              - in: query
                name: id
                schema:
                  type: string
            responses:
              '200':
                content:
                  application/json:
                    schema:
                      items:
                        format: int32
                        type: integer
                      type: array
                description: a response description
        """
    )


def test_multiple(convert_paths, oas_fragment):
    converted = convert_paths(
        oas_fragment(
            """
            /test:
              get:
                responses:
                  '200':
                    description: a test response description
            /eggs:
              post:
                responses:
                  '201':
                    description: an eggs response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        /test:
          get:
            responses:
              '200':
                description: a test response description
        /eggs:
          post:
            responses:
              '201':
                description: an eggs response description
        """
    )


def test_vendor_extensions(convert_paths, oas_fragment):
    converted = convert_paths(
        oas_fragment(
            """
            /test:
              get:
                responses:
                  '200':
                    description: a response description
            x-vendor-ext: vendor-ext
            """
        ),
    )
    assert converted == oas_fragment(
        """
        /test:
          get:
            responses:
              '200':
                description: a response description
        x-vendor-ext: vendor-ext
        """
    )



import textwrap

import pytest
import yaml


@pytest.fixture(scope="function")
def oas_fragment():
    def oas_fragment(fragment):
        return yaml.safe_load(textwrap.dedent(fragment))

    return oas_fragment



""".convert_response() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_response(oas_fragment):
    def _wrapper(response, produces):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"]["responses"]["200"] = response
        oas2["paths"]["/test"]["get"]["produces"] = produces

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"]["responses"]["200"]

    return _wrapper


def test_minimal(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        """
    )


def test_schema(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            schema:
              items:
                format: int32
                type: integer
              type: array
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            schema:
              items:
                format: int32
                type: integer
              type: array
        description: a response description
        """
    )


def test_schema_mimetypes(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            schema:
              items:
                format: int32
                type: integer
              type: array
            """
        ),
        produces=["application/json", "text/plain"],
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            schema:
              items:
                format: int32
                type: integer
              type: array
          text/plain:
            schema:
              items:
                format: int32
                type: integer
              type: array
        description: a response description
        """
    )


def test_schema_no_mimetypes(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            schema:
              items:
                format: int32
                type: integer
              type: array
            """
        ),
        produces=None,
    )
    assert converted == oas_fragment(
        """
        content:
          '*/*':
            schema:
              items:
                format: int32
                type: integer
              type: array
        description: a response description
        """
    )


def test_examples(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            examples:
              application/json:
                something: important
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            example:
              something: important
        description: a response description
        """
    )


def test_examples_any_type(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            examples:
              application/json: '{"something": "important"}'
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            example: '{"something": "important"}'
        description: a response description
        """
    )


def test_examples_mimetypes(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            examples:
              application/json:
                something: important
              text/plain: something=imporant
            """
        ),
        produces=["application/json", "text/plain"],
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            example:
              something: important
          text/plain:
            example: something=imporant
        description: a response description
        """
    )


def test_headers_schema_only(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            headers:
              X-Test:
                type: string
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        headers:
          X-Test:
            schema:
              type: string
        """
    )


def test_headers_schema_extra(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            headers:
              X-Test:
                description: Is it a test?
                type: string
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        headers:
          X-Test:
            description: Is it a test?
            schema:
              type: string
        """
    )


def test_headers_multiple(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            headers:
              X-Bar:
                format: int32
                type: integer
              X-Foo:
                type: string
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        headers:
          X-Bar:
            schema:
              format: int32
              type: integer
          X-Foo:
            schema:
              type: string
        """
    )


def test_schema_examples_headers(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            examples:
              application/json:
                something: important
            headers:
              X-Test:
                description: Is it a test?
                type: string
            schema:
              items:
                format: int32
                type: integer
              type: array
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        content:
          application/json:
            example:
              something: important
            schema:
              items:
                format: int32
                type: integer
              type: array
        headers:
          X-Test:
            description: Is it a test?
            schema:
              type: string
        """
    )


def test_complete(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            examples:
              application/json:
                something: important
            headers:
              X-Test:
                description: Is it a test?
                type: string
            schema:
              items:
                format: int32
                type: integer
              type: array
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        content:
          application/json:
            example:
              something: important
            schema:
              items:
                format: int32
                type: integer
              type: array
        headers:
          X-Test:
            description: Is it a test?
            schema:
              type: string
        """
    )


def test_vendor_extensions(convert_response, oas_fragment):
    converted = convert_response(
        oas_fragment(
            """
            description: a response description
            examples:
              application/json:
                something: important
            headers:
              X-Test:
                description: Is it a test?
                type: string
                x-header-ext: header-ext
            schema:
              items:
                format: int32
                type: integer
              type: array
              x-schema-ext: schema-ext
            x-response-ext: response-ext
            """
        ),
        produces=["application/json"],
    )
    assert converted == oas_fragment(
        """
        description: a response description
        content:
          application/json:
            example:
              something: important
            schema:
              items:
                format: int32
                type: integer
              type: array
              x-schema-ext: schema-ext
        headers:
          X-Test:
            description: Is it a test?
            schema:
              type: string
            x-header-ext: header-ext
        x-response-ext: response-ext
        """
    )



"""
    sphinxcontrib.openapi.openapi30
    -------------------------------

    The OpenAPI 3.0.0 spec renderer. Based on ``sphinxcontrib-httpdomain``.

    :copyright: (c) 2016, Ihor Kalnytskyi.
    :license: BSD, see LICENSE for details.
"""

import copy

import collections
import collections.abc
import textwrap

from datetime import datetime
import itertools
import json
import re
from urllib import parse
from http.client import responses as http_status_codes
from sphinxcontrib.openapi.renderers._model import _process_one, _entities

from sphinx.util import logging
from sphinx.locale import get_translation

from sphinxcontrib.openapi import utils

_ = get_translation('openapi')


LOG = logging.getLogger(__name__)

# https://github.com/OAI/OpenAPI-Specification/blob/3.0.2/versions/3.0.0.md#data-types
_TYPE_MAPPING = {
    ('integer', 'int32'): 1,  # integer
    ('integer', 'int64'): 1,  # long
    ('number', 'float'): 1.0,  # float
    ('number', 'double'): 1.0,  # double
    ('boolean', None): True,  # boolean
    ('string', None): 'string',  # string
    ('string', 'byte'): 'c3RyaW5n',  # b'string' encoded in base64,  # byte
    ('string', 'binary'): '01010101',  # binary
    ('string', 'date'): datetime.now().date().isoformat(),  # date
    ('string', 'date-time'): datetime.now().isoformat(),  # dateTime
    ('string', 'password'): '********',  # password

    # custom extensions to handle common formats
    ('string', 'email'): 'name@example.com',
    ('string', 'zip-code'): '90210',
    ('string', 'uri'): 'https://example.com',

    # additional fallthrough cases
    ('integer', None): 1,  # integer
    ('number', None): 1.0,  # <fallthrough>
}

_READONLY_PROPERTY = object()  # sentinel for values not included in requests


def _dict_merge(dct, merge_dct):
    """Recursive dict merge.

    Inspired by :meth:``dict.update()``, instead of updating only top-level
    keys, dict_merge recurses down into dicts nested to an arbitrary depth,
    updating keys. The ``merge_dct`` is merged into ``dct``.

    From https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

    Arguments:
        dct: dict onto which the merge is executed
        merge_dct: dct merged into dct
    """
    for k in merge_dct.keys():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], collections.abc.Mapping)):
            _dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def _parse_schema(schema, method):
    """
    Convert a Schema Object to a Python object.

    Args:
        schema: An ``OrderedDict`` representing the schema object.
    """
    if method and schema.get('readOnly', False):
        return _READONLY_PROPERTY

    # allOf: Must be valid against all of the subschemas
    if 'allOf' in schema:
        schema_ = copy.deepcopy(schema['allOf'][0])
        for x in schema['allOf'][1:]:
            _dict_merge(schema_, x)

        return _parse_schema(schema_, method)

    # anyOf: Must be valid against any of the subschemas
    # TODO(stephenfin): Handle anyOf

    # oneOf: Must be valid against exactly one of the subschemas
    if 'oneOf' in schema:
        # we only show the first one since we can't show everything
        return _parse_schema(schema['oneOf'][0], method)

    if 'enum' in schema:
        if 'example' in schema:
            return schema['example']
        # we show the first one
        return schema['enum'][0]

    schema_type = schema.get('type', 'object')

    if schema_type == 'array':
        # special case oneOf and anyOf so that we can show examples for all
        # possible combinations
        if 'oneOf' in schema['items']:
            return [
                _parse_schema(x, method) for x in schema['items']['oneOf']
            ]

        if 'anyOf' in schema['items']:
            return [
                _parse_schema(x, method) for x in schema['items']['anyOf']
            ]

        return [_parse_schema(schema['items'], method)]

    if schema_type == 'object':
        if 'example' in schema:
            example = schema.get('example')
            if example:
                example = copy.deepcopy(example)
                # filters out readonly properties
                if method and 'properties' in schema:
                    for k, v in schema.get('properties', {}).items():
                        if v.get('readOnly', False) and k in example:
                            del example[k]
                ret = collections.OrderedDict(example)
                # XXX should be True to be compliant with OpenAPI
                if schema.get('additionalProperties', False) and '...' not in example:
                    # materialize in the example the fact that additional properties can be added
                    ret['...'] = '...'
                return ret
        if method and 'properties' in schema and \
                all(v.get('readOnly', False)
                    for v in schema['properties'].values()):
            return _READONLY_PROPERTY

        results = []
        for name, prop in schema.get('properties', {}).items():
            result = _parse_schema(prop, method)
            if result != _READONLY_PROPERTY:
                results.append((name, result))

        # XXX should be True to be compliant with OpenAPI
        if schema.get('additionalProperties', False):
            # materialize in the example the fact that additional properties can be added
            results.append(("...", "..."))

        return collections.OrderedDict(results)

    if 'example' in schema:
        return schema['example']
    if (schema_type, schema.get('format')) in _TYPE_MAPPING:
        return _TYPE_MAPPING[(schema_type, schema.get('format'))]

    return _TYPE_MAPPING[(schema_type, None)]  # unrecognized format


def _example(media_type_objects, method=None, endpoint=None, status=None,
             reqheader_examples={},
             resheader_examples={},
             nb_indent=0):
    """
    Format examples in `Media Type Object` openapi v3 to HTTP request or
    HTTP response example.
    If method and endpoint is provided, this function prints a request example
    else status should be provided to print a response example.

    Arguments:
        media_type_objects (Dict[str, Dict]): Dict containing
            Media Type Objects.
        method: The HTTP method to use in example.
        endpoint: The HTTP route to use in example.
        status: The HTTP status to use in example.
    """
    indent = '   '
    extra_indent = indent * nb_indent

    if method is not None:
        method = method.upper()
    else:
        try:
            # one of possible values for status might be 'default'.
            # in the case, just fallback to '-'
            status_text = http_status_codes[int(status)]
        except (ValueError, KeyError):
            status_text = '-'

    # Provide request samples for GET requests
    if method == 'GET':
        media_type_objects[''] = {
            'examples': {_('Example request'): {'value': '\n'}}}

    for content_type, content in media_type_objects.items():
        examples = content.get('examples')
        example = content.get('example')

        # Try to get the example from the schema
        if example is None and 'schema' in content:
            example = content['schema'].get('example')
            if example:
                example = copy.deepcopy(example)
                # filters out readonly properties
                if method and 'properties' in content['schema']:
                    for k, v in content['schema'].get('properties', {}).items():
                        if v.get('readOnly', False) and k in example:
                            del example[k]
                # XXX should be True to be compliant with OpenAPI
                if content['schema'].get('additionalProperties', False) and '...' not in example:
                    # materialize in the example the fact that additional properties can be added
                    example['...'] = '...'

        if examples is None:
            examples = {}
            if not example:
                if re.match(r"application/[a-zA-Z\+]*json", content_type) is \
                        None:
                    LOG.info('skipping non-JSON example generation.')
                    continue
                example = _parse_schema(content['schema'], method=method)

            if method is None:
                examples[_('Example response')] = {
                    'value': example,
                }
            else:
                examples[_('Example request')] = {
                    'value': example,
                }

        for example in examples.values():
            # According to OpenAPI v3 specs, string examples should be left unchanged
            if not isinstance(example['value'], str):
                example['value'] = json.dumps(
                    example['value'], indent=4, separators=(',', ': '))

        for example_name, example in examples.items():
            if 'summary' in example:
                example_title = '{example_name} - {example[summary]}'.format(
                    **locals())
            else:
                example_title = example_name

            yield ''
            yield '{extra_indent}**{example_title}:**'.format(**locals())
            yield ''
            yield '{extra_indent}.. sourcecode:: http'.format(**locals())
            yield ''

            # Print http request example
            if method:
                yield '{extra_indent}{indent}{method} {endpoint} HTTP/1.1' \
                    .format(**locals())
                yield '{extra_indent}{indent}Host: example.com' \
                    .format(**locals())
                if content_type:
                    yield '{extra_indent}{indent}Content-Type: {content_type}'\
                        .format(**locals())
                for k, v in reqheader_examples.items():
                    yield '{extra_indent}{indent}{k}: {v}'\
                        .format(**locals())
            # Print http response example
            else:
                yield '{extra_indent}{indent}HTTP/1.1 {status} {status_text}' \
                    .format(**locals())
                yield '{extra_indent}{indent}Content-Type: {content_type}' \
                    .format(**locals())
                for k, v in resheader_examples.items():
                    yield '{extra_indent}{indent}{k}: {v}'\
                        .format(**locals())

            if content_type:
                yield ''
                for example_line in example['value'].splitlines():
                    yield '{extra_indent}{indent}{example_line}'.format(
                        **locals()
                        )
            if example['value'].splitlines():
                yield ''


def ref2link(entities, ref):
    name = ref.split('/')[-1]
    ref = entities(ref)
    return ':ref:`{name} <{ref}>`'.format(**locals())


def _httpresource(endpoint, method, properties, convert, render_examples,
                  render_request, group_examples=False, entities=False):
    # https://github.com/OAI/OpenAPI-Specification/blob/3.0.2/versions/3.0.0.md#operation-object
    parameters = properties.get('parameters', [])
    responses = properties['responses']
    query_param_examples = []
    indent = '   '

    yield '.. http:{0}:: {1}'.format(method, endpoint)
    yield '   :synopsis: {0}'.format(properties.get('summary', 'null'))
    yield ''

    if 'summary' in properties:
        for line in properties['summary'].splitlines():
            yield '{indent}**{line}**'.format(**locals())
        yield ''

    if 'description' in properties:
        for line in convert(properties['description']).splitlines():
            yield '{indent}{line}'.format(**locals())
        yield ''

    if properties.get('deprecated', False):
        yield '{indent}**DEPRECATED**'.format(**locals())
        yield ''

    if 'security' in properties:
        for sec_schema in properties['security']:
            sec_scope = ' or '.join([
                '``{}``'.format(s) for sch in sec_schema.values() for s in sch
            ])
            s = '{indent}**' + _('Scope required') + '**: {sec_scope}'
            yield s.format(**locals())
        yield ''

    def get_desc(desc, schema, indent, deep=True):
        if entities:
            doc = next(_process_one(['R'], schema, False, entities, convert))
            if desc:
                if not desc[-1] == '.':
                    desc = desc + '.'
            if doc[1]:
                if not doc[1].startswith(_("Object of")) and not doc[1].startswith(_("Array of")):
                    doc[1] = _("Object of type {}").format(doc[1])
                if not doc[1][-1] == '.':
                    doc[1] = doc[1] + '.'
                desc += '\n' + doc[1]
            if deep and doc[2] and doc[2] != _('Additional properties'):
                if not doc[2][-1] == '.':
                    doc[2] = doc[2] + '.'
                desc += '\n' + doc[2]
            desc = desc.rstrip()
        else:
            doc = next(_process_one(['R'], schema, False, entities, convert))
            if desc:
                if not desc[-1] == '.':
                    desc = desc + '.'
            if doc[2] and doc[2] != _('Additional properties'):
                if not doc[2][-1] == '.':
                    doc[2] = doc[2] + '.'
                desc += '\n' + doc[2]
            desc = desc.rstrip()
        desc = textwrap.indent(desc, '{indent}{indent}'.format(**locals())).lstrip()
        return desc

    # print request's path params
    for param in filter(lambda p: p['in'] == 'path', parameters):
        yield indent + ':param {type} {name}:'.format(
            type=param['schema']['type'],
            name=param['name'])

        desc = param.get('description', '')
        desc = get_desc(desc, param['schema'], indent)
        if desc:
            yield '{indent}{indent}{desc}'.format(**locals())

    # print request's query params
    for param in filter(lambda p: p['in'] == 'query', parameters):
        yield indent + ':query {type} {name}:'.format(
            type=param['schema']['type'],
            name=param['name'])
        desc = param.get('description', '')
        desc = get_desc(desc, param['schema'], indent)
        if desc:
            yield '{indent}{indent}{desc}'.format(**locals())
        if param.get('required', False):
            yield '{indent}{indent}'.format(**locals()) + \
                '({})'.format(_('Required'))
        example = _parse_schema(param['schema'], method)
        example = param.get('example', example)
        if param.get('explode', False) and isinstance(example, list):
            for v in example:
                query_param_examples.append((param['name'], v))
        elif param.get('explode', False) and isinstance(example, dict):
            for k, v in example.items():
                query_param_examples.append((k, v))
        else:
            query_param_examples.append((param['name'], example))

    # print request content
    if render_request:
        request_content = properties.get('requestBody', {}).get('content', {})
        if request_content and 'application/json' in request_content:
            schema = request_content['application/json']['schema']
            req_properties = json.dumps(schema['properties'], indent=2,
                                        separators=(',', ':'))
            yield '{indent}'.format(**locals()) + '**{}**'.format(_('Request body:'))
            yield ''
            yield '{indent}.. sourcecode:: json'.format(**locals())
            yield ''
            for line in req_properties.splitlines():
                # yield indent + line
                yield '{indent}{indent}{line}'.format(**locals())
                # yield ''
    else:
        desc = properties.get('requestBody', {}).get('description', '')
        request_content = properties.get('requestBody', {}).get('content', {})
        if request_content and 'application/json' in request_content:
            schema = request_content['application/json'].get('schema', {})
            if '$entity_ref' in schema or schema.get('type', 'object') == 'array':
                desc = get_desc(desc, schema, indent, deep=False)
                if desc:
                    yield '{indent}:form body: {desc}'.format(**locals())
            else:
                for prop, v in schema.get('properties', {}).items():
                    ptype = v.get('type', '')
                    desc = v.get('description', '')
                    yield '{indent}:jsonparam {ptype} {prop}: {desc}'.format(**locals()).rstrip()
        else:
            if desc:
                yield '{indent}:form body: {desc}.'.format(**locals())

    # print request header params
    reqheader_examples = {}
    for param in filter(lambda p: p['in'] == 'header', parameters):
        yield indent + ':reqheader {name}:'.format(**param)
        desc = param.get('description', '')
        desc = get_desc(desc, param['schema'], indent)
        if desc:
            yield '{indent}{indent}{desc}'.format(**locals())
        if param.get('required', False):
            yield '{indent}{indent}(Required)'.format(**locals())
        ex = param.get('example', param.get('schema', {}).get('example', None))
        if ex is None:
            # try examples
            ex = param.get('examples', param.get('schema', {}).get('examples', [None]))[0]
        if ex:
            reqheader_examples[param['name']] = ex

    # print request example
    if render_examples and not group_examples:
        endpoint_examples = endpoint
        if query_param_examples:
            endpoint_examples = endpoint + "?" + \
                parse.urlencode(query_param_examples)

        # print request example
        request_content = properties.get('requestBody', {}).get('content', {})
        for line in _example(
                request_content,
                method,
                endpoint=endpoint_examples,
                reqheader_examples=reqheader_examples,
                nb_indent=1):
            yield line

    # print response headers
    resheader_examples = {}
    for status, response in responses.items():
        for headername, header in response.get('headers', {}).items():
            yield indent + ':resheader {name}:'.format(name=headername)
            desc = header.get('description', '')
            desc = get_desc(desc, header.get('schema', {}), indent)
            if desc:
                yield '{indent}{indent}{desc}'.format(**locals())
            ex = header.get('example', header.get('schema', {}).get('example', None))
            if ex is None:
                # try examples
                ex = header.get('examples', header.get('schema', {}).get('examples', [None]))[0]
            if ex:
                resheader_examples[headername] = ex

    # print response status codes
    for status, response in responses.items():
        yield '{indent}:status {status}:'.format(**locals())
        content = response.get('content', {})
        if entities and content and 'application/json' in content:
            schema = content['application/json']['schema']
            desc = response.get('description', '')
            desc = get_desc(desc, schema, indent, deep=False)
            if desc:
                yield '{indent}{indent}{desc}'.format(**locals())
        else:
            desc = response.get('description', '')
            if desc and desc[-1] != '.':
                desc += '.'
            for line in convert(desc.splitlines()):
                yield '{indent}{indent}{line}'.format(**locals())

        # print response example
        if render_examples and not group_examples:
            for line in _example(
                    response.get('content', {}),
                    status=status,
                    resheader_examples=resheader_examples,
                    nb_indent=2):
                yield line

    if render_examples and group_examples:
        endpoint_examples = endpoint
        if query_param_examples:
            endpoint_examples = endpoint + "?" + \
                parse.urlencode(query_param_examples)

        # print request example
        request_content = properties.get('requestBody', {}).get('content', {})
        for line in _example(
                request_content,
                method,
                endpoint=endpoint_examples,
                reqheader_examples=reqheader_examples,
                resheader_examples=resheader_examples,
                nb_indent=1):
            yield line

        # print response example
        for status, response in responses.items():
            for line in _example(
                    response.get('content', {}),
                    status=status,
                    reqheader_examples=reqheader_examples,
                    resheader_examples=resheader_examples,
                    nb_indent=1):
                yield line

    for cb_name, cb_specs in properties.get('callbacks', {}).items():
        yield ''
        yield indent + '.. admonition:: Callback: ' + cb_name
        yield ''

        for cb_endpoint in cb_specs.keys():
            for cb_method, cb_properties in cb_specs[cb_endpoint].items():
                for line in _httpresource(
                        cb_endpoint,
                        cb_method,
                        cb_properties,
                        convert=convert,
                        render_examples=render_examples,
                        render_request=render_request,
                        group_examples=group_examples,
                        entities=entities):
                    if line:
                        yield indent+indent+line
                    else:
                        yield ''

    yield ''


def _header(title):
    yield title
    yield '=' * len(title)
    yield ''


def openapihttpdomain(spec, **options):
    generators = []

    # OpenAPI spec may contain JSON references, common properties, etc.
    # Trying to render the spec "As Is" will require to put multiple
    # if-s around the code. In order to simplify flow, let's make the
    # spec to have only one (expected) schema, i.e. normalize it.
    utils.normalize_spec(spec, **options)

    # Paths list to be processed
    paths = []

    # If 'paths' are passed we've got to ensure they exist within an OpenAPI
    # spec; otherwise raise error and ask user to fix that.
    if 'paths' in options:
        if not set(options['paths']).issubset(spec['paths']):
            raise ValueError(
                'One or more paths are not defined in the spec: %s.' % (
                    ', '.join(set(options['paths']) - set(spec['paths'])),
                )
            )
        paths = options['paths']

    # Check against regular expressions to be included
    if 'include' in options:
        for i in options['include']:
            ir = re.compile(i)
            for path in spec['paths']:
                if ir.match(path):
                    paths.append(path)

    # If no include nor paths option, then take full path
    if 'include' not in options and 'paths' not in options:
        paths = spec['paths']

    # Remove paths matching regexp
    if 'exclude' in options:
        _paths = []
        for e in options['exclude']:
            er = re.compile(e)
            for path in paths:
                if not er.match(path):
                    _paths.append(path)
        paths = _paths

    render_request = False
    if 'request' in options:
        render_request = True

    convert = utils.get_text_converter(options)

    if 'entities' in options:
        def f_entities(x):
            return _entities(spec, x)
        entities = f_entities
    else:
        entities = False

    # https://github.com/OAI/OpenAPI-Specification/blob/3.0.2/versions/3.0.0.md#paths-object
    if 'group' in options:
        groups = collections.OrderedDict(
            [(x['name'], []) for x in spec.get('tags', {})]
            )

        for endpoint in paths:
            for method, properties in spec['paths'][endpoint].items():
                key = properties.get('tags', [''])[0]
                groups.setdefault(key, []).append(_httpresource(
                    endpoint,
                    method,
                    properties,
                    convert,
                    render_examples='examples' in options,
                    render_request=render_request,
                    group_examples='group_examples' in options,
                    entities=entities))

        for key in groups.keys():
            if key:
                generators.append(_header(key))
            else:
                generators.append(_header('default'))

            generators.extend(groups[key])
    else:
        for endpoint in paths:
            for method, properties in spec['paths'][endpoint].items():
                generators.append(_httpresource(
                    endpoint,
                    method,
                    properties,
                    convert,
                    render_examples='examples' in options,
                    render_request=render_request,
                    group_examples='group_examples' in options,
                    entities=entities))

    return iter(itertools.chain(*generators))



"""
    sphinxcontrib.openapi.utils
    ---------------------------

    Common functionality shared across the various renderers.

    :copyright: (c) 2016, Ihor Kalnytskyi.
    :license: BSD, see LICENSE for details.
"""

from __future__ import unicode_literals

import collections
import collections.abc

from contextlib import closing
import jsonschema
import yaml
try:
    from m2r import convert as convert_markdown
except ImportError:
    convert_markdown = None

from urllib.parse import urlsplit
from urllib.request import urlopen

import os.path


class OpenApiRefResolver(jsonschema.RefResolver):
    """
    Overrides resolve_remote to support both YAML and JSON
    OpenAPI schemas.
    """

    try:
        import requests
        _requests = requests
    except ImportError:
        _requests = None

    def resolve_remote(self, uri):
        scheme, _, path, _, _ = urlsplit(uri)
        _, extension = os.path.splitext(path)

        if extension not in [".yml", ".yaml"] or scheme in self.handlers:
            return super(OpenApiRefResolver, self).resolve_remote(uri)

        if scheme in [u"http", u"https"] and self._requests:
            response = self._requests.get(uri)
            result = yaml.safe_load(response)
        else:
            # Otherwise, pass off to urllib and assume utf-8
            with closing(urlopen(uri)) as url:
                response = url.read().decode("utf-8")
                result = yaml.safe_load(response)

        if self.cache_remote:
            self.store[uri] = result
        return result


def _resolve_refs(uri, spec):
    """Resolve JSON references in a given dictionary.

    OpenAPI spec may contain JSON references to its nodes or external
    sources, so any attempt to rely that there's some expected attribute
    in the spec may fail. So we need to resolve JSON references before
    we use it (i.e. replace with referenced object). For details see:

        https://tools.ietf.org/html/draft-pbryan-zyp-json-ref-02

    The input spec is modified in-place despite being returned from
    the function.
    """

    resolver = OpenApiRefResolver(uri, spec)

    def _do_resolve(node):
        if isinstance(node, collections.abc.Mapping) and '$ref' in node:
            ref = node['$ref']
            with resolver.resolving(node['$ref']) as resolved:
                ret = _do_resolve(resolved)  # might have recursive references
            # Restore the $ref in case we want to
            # separate the entities in the document
            if isinstance(ret, collections.abc.Mapping):
                ret['$entity_ref'] = ref
            return ret
        elif isinstance(node, collections.abc.Mapping):
            for k, v in node.items():
                node[k] = _do_resolve(v)
        elif isinstance(node, (list, tuple)):
            for i in range(len(node)):
                node[i] = _do_resolve(node[i])
        return node

    return _do_resolve(spec)


def normalize_spec(spec, **options):
    # OpenAPI spec may contain JSON references, so we need resolve them
    # before we access the actual values trying to build an httpdomain
    # markup. Since JSON references may be relative, it's crucial to
    # pass a document URI in order to properly resolve them.
    spec = _resolve_refs(options.get('uri', ''), spec)

    # OpenAPI spec may contain common endpoint's parameters top-level.
    # In order to do not place if-s around the code to handle special
    # cases, let's normalize the spec and push common parameters inside
    # endpoints definitions.
    for endpoint in spec['paths'].values():
        parameters = endpoint.pop('parameters', [])
        for method in endpoint.values():
            method.setdefault('parameters', [])
            method['parameters'].extend(parameters)


def get_text_converter(options):
    """Decide on a text converter for prose."""
    if 'format' in options:
        if options['format'] == 'markdown':
            if convert_markdown is None:
                raise ValueError(
                    "Markdown conversion isn't available, "
                    "install the [markdown] extra."
                )
            return convert_markdown

    # No conversion needed.
    return lambda s: s



"""
    sphinxcontrib
    ~~~~~~~~~~~~~

    This package is a namespace package that contains all extensions
    distributed in the ``sphinx-contrib`` distribution.

    :copyright: (c) 2016 by Igor Kalnitsky.
    :license: BSD, see LICENSE for details.
"""

__import__("pkg_resources").declare_namespace(__name__)



""".convert_parameters() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


_MISSING = object()


@pytest.fixture(scope="function")
def convert_parameters(oas_fragment):
    def _wrapper(parameters):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"]["parameters"] = parameters

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"].get("parameters", _MISSING)

    return _wrapper


def test_header_path_query(convert_parameters, oas_fragment):
    converted = convert_parameters(
        oas_fragment(
            """
            - in: header
              name: token
              type: string
            - in: path
              name: username
              required: true
              type: string
            - in: query
              name: id
              type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        - in: header
          name: token
          schema:
            type: string
        - in: path
          name: username
          required: true
          schema:
            type: string
        - in: query
          name: id
          schema:
            type: string
        """
    )


def test_body_is_ignored(convert_parameters, oas_fragment):
    converted = convert_parameters(
        oas_fragment(
            """
            - description: user to add to the system
              in: body
              name: user
              required: true
              schema:
                $ref: '#/definitions/User'
            """
        ),
    )
    assert converted is _MISSING


def test_formData_is_ignored(convert_parameters, oas_fragment):
    converted = convert_parameters(
        oas_fragment(
            """
            - description: The avatar of the user
              in: formData
              name: avatar
              type: file
            """
        ),
    )

    assert converted is _MISSING



"""OpenAPI spec renderer: render_operation."""

import textwrap

import pytest

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


def test_render_operation(testrenderer, oas_fragment):
    """Usual operation definition is rendered."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences/{evidenceId}",
            "get",
            oas_fragment(
                """
                summary: Retrieve an evidence by ID.
                description: More verbose description...
                parameters:
                  - name: evidenceId
                    in: path
                    required: true
                    description: A unique evidence identifier to query.
                    schema:
                      type: string
                  - name: details
                    in: query
                    description: If true, information w/ details is returned.
                    schema:
                      type: boolean
                responses:
                  '200':
                    description: An evidence.
                  '404':
                    description: An evidence not found.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /evidences/{evidenceId}

           **Retrieve an evidence by ID.**

           More verbose description...

           :param evidenceId:
              A unique evidence identifier to query.
           :paramtype evidenceId: string, required
           :queryparam details:
              If true, information w/ details is returned.
           :queryparamtype details: boolean
           :statuscode 200:
              An evidence.
           :statuscode 404:
              An evidence not found.
        """.rstrip()
    )


def test_render_operation_minimal(testrenderer, oas_fragment):
    """Operation minimal definition is rendered."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_summary(testrenderer, oas_fragment):
    """Operation's 'summary' is rendered in bold."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                summary: Create an evidence.
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           **Create an evidence.**

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_description(testrenderer, oas_fragment):
    """Operation's 'description' is rendered."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                description: Create an evidence.
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           Create an evidence.

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_description_multiline(testrenderer, oas_fragment):
    """Operation's multiline 'description' is rendered."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                description: |
                  Create
                  an evidence.
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           Create
           an evidence.

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_description_commonmark_default(testrenderer, oas_fragment):
    """Operation's 'description' must be in commonmark by default."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                description: __Create__ an `evidence`.
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           **Create** an ``evidence``.

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_description_commonmark(fakestate, oas_fragment):
    """Operation's 'description' can be in commonmark."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                description: __Create__ an `evidence`.
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           **Create** an ``evidence``.

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_description_commonmark_restructuredtext(
    fakestate, oas_fragment
):
    """Operation's 'description' can be in restructuredtext."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                description: __Create__ an `evidence`.
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           __Create__ an `evidence`.

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_deprecated(testrenderer, oas_fragment):
    """Operation's 'deprecated' mark is rendered."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                responses:
                  '201':
                    description: An evidence created.
                deprecated: true
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences
           :deprecated:

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


def test_render_operation_w_requestbody(testrenderer, oas_fragment):
    """Operation's 'requestBody' is rendered."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            "post",
            oas_fragment(
                """
                requestBody:
                  content:
                    application/json:
                      example:
                        foo: bar
                        baz: 42
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:post:: /evidences

           .. sourcecode:: http

              POST /evidences HTTP/1.1
              Content-Type: application/json

              {
                "foo": "bar",
                "baz": 42
              }

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["method"], [pytest.param("POST"), pytest.param("pOst"), pytest.param("post")]
)
def test_render_operation_caseinsensitive_method(testrenderer, method, oas_fragment):
    """Operation's 'method' is case insensitive."""

    markup = textify(
        testrenderer.render_operation(
            "/evidences",
            method,
            oas_fragment(
                """
                requestBody:
                  content:
                    application/json:
                      example:
                        foo: bar
                        baz: 42
                responses:
                  '201':
                    description: An evidence created.
                """
            ),
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        .. http:{method}:: /evidences

           .. sourcecode:: http

              {method.upper()} /evidences HTTP/1.1
              Content-Type: application/json

              {{
                "foo": "bar",
                "baz": 42
              }}

           :statuscode 201:
              An evidence created.
        """.rstrip()
    )



"""
    sphinxcontrib.openapi.directive
    -------------------------------

    The main directive for the extension.

    :copyright: (c) 2016, Ihor Kalnytskyi.
    :license: BSD, see LICENSE for details.
"""

import functools

from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective
import yaml


# Locally cache spec to speedup processing of same spec file in multiple
# openapi directives
@functools.lru_cache()
def _get_spec(abspath, encoding):
    with open(abspath, 'rt', encoding=encoding) as stream:
        return yaml.safe_load(stream)


def create_directive_from_renderer(renderer_cls):
    """Create rendering directive from a renderer class."""

    class _RenderingDirective(SphinxDirective):
        required_arguments = 1                  # path to openapi spec
        final_argument_whitespace = True        # path may contain whitespaces
        option_spec = dict(
            {
                'encoding': directives.encoding,    # useful for non-ascii cases :)
            },
            **renderer_cls.option_spec
        )

        def run(self):
            relpath, abspath = self.env.relfn2path(directives.path(self.arguments[0]))

            # URI parameter is crucial for resolving relative references. So we
            # need to set this option properly as it's used later down the
            # stack.
            self.options.setdefault('uri', 'file://%s' % abspath)

            # Add a given OpenAPI spec as a dependency of the referring
            # reStructuredText document, so the document is rebuilt each time
            # the spec is changed.
            self.env.note_dependency(relpath)

            # Read the spec using encoding passed to the directive or fallback to
            # the one specified in Sphinx's config.
            encoding = self.options.get('encoding', self.config.source_encoding)
            spec = _get_spec(abspath, encoding)
            return renderer_cls(self.state, self.options).render(spec)

    return _RenderingDirective



"""
    sphinxcontrib.openapi.openapi20
    -------------------------------

    The OpenAPI 2.0 (f.k.a. Swagger) spec renderer. Based on
    ``sphinxcontrib-httpdomain``.

    :copyright: (c) 2016, Ihor Kalnytskyi.
    :license: BSD, see LICENSE for details.
"""

import collections
import itertools
import re

from sphinxcontrib.openapi import utils


def _httpresource(endpoint, method, properties, convert):
    parameters = properties.get('parameters', [])
    responses = properties['responses']
    indent = '   '

    yield '.. http:{0}:: {1}'.format(method, endpoint)
    yield '   :synopsis: {0}'.format(properties.get('summary', 'null'))
    yield ''

    if 'summary' in properties:
        for line in properties['summary'].splitlines():
            yield '{indent}**{line}**'.format(**locals())
        yield ''

    if 'description' in properties:
        for line in convert(properties['description']).splitlines():
            yield '{indent}{line}'.format(**locals())
        yield ''

    for param in filter(lambda p: p['in'] == 'path', parameters):
        yield indent + ':param {type} {name}:'.format(**param)
        for line in convert(param.get('description', '')).splitlines():
            yield '{indent}{indent}{line}'.format(**locals())

    # print request's query params
    for param in filter(lambda p: p['in'] == 'query', parameters):
        yield indent + ':query {type} {name}:'.format(**param)
        for line in convert(param.get('description', '')).splitlines():
            yield '{indent}{indent}{line}'.format(**locals())

    # print the json body params
    for param in filter(lambda p: p['in'] == 'body', parameters):
        if 'schema' in param:
            yield ''
            for line in convert_json_schema(param['schema']):
                yield '{indent}{line}'.format(**locals())
            yield ''

    # print response status codes
    for status, response in sorted(responses.items()):
        yield '{indent}:status {status}:'.format(**locals())
        for line in convert(response['description']).splitlines():
            yield '{indent}{indent}{line}'.format(**locals())

    # print request header params
    for param in filter(lambda p: p['in'] == 'header', parameters):
        yield indent + ':reqheader {name}:'.format(**param)
        for line in convert(param.get('description', '')).splitlines():
            yield '{indent}{indent}{line}'.format(**locals())

    # print response headers
    for status, response in responses.items():
        for headername, header in response.get('headers', {}).items():
            yield indent + ':resheader {name}:'.format(name=headername)
            for line in convert(header['description']).splitlines():
                yield '{indent}{indent}{line}'.format(**locals())

    for status, response in responses.items():
        if not is_2xx_response(status):
            continue
        if 'schema' in response:
            yield ''
            for line in convert_json_schema(
                    response['schema'], directive=':>json'):
                yield '{indent}{line}'.format(**locals())
            yield ''

    yield ''


def convert_json_schema(schema, directive=':<json'):
    """
    Convert json schema to `:<json` sphinx httpdomain.
    """

    output = []

    def _convert(schema, name='', required=False):
        """
        Fill the output list, with 2-tuple (name, template)

        i.e: ('user.age', 'str user.age: the age of user')

        This allow to sort output by field name
        """

        type_ = schema.get('type', 'any')
        required_properties = schema.get('required', ())
        if type_ == 'object' and schema.get('properties'):
            for prop, next_schema in schema.get('properties', {}).items():
                _convert(
                    next_schema, '{name}.{prop}'.format(**locals()),
                    (prop in required_properties))

        elif type_ == 'array':
            _convert(schema['items'], name + '[]')

        else:
            if name:
                name = name.lstrip('.')
                constraints = []
                if required:
                    constraints.append('required')
                if schema.get('readOnly', False):
                    constraints.append('read only')
                if constraints:
                    constraints = '({})'.format(', '.join(constraints))
                else:
                    constraints = ''

                if schema.get('description', ''):
                    if constraints:
                        output.append((
                            name,
                            '{type_} {name}:'
                            ' {schema[description]}'
                            ' {constraints}'.format(**locals())))
                    else:
                        output.append((
                            name,
                            '{type_} {name}:'
                            ' {schema[description]}'.format(**locals())))

                else:
                    if constraints:
                        output.append(
                            (name,
                             '{type_} {name}:'
                             ' {constraints}'.format(**locals())))
                    else:
                        output.append(
                            (name,
                             '{type_} {name}:'.format(**locals())))

    _convert(schema)

    for _, render in sorted(output):
        yield '{} {}'.format(directive, render)


def is_2xx_response(status):
    try:
        status = int(status)
        return 200 <= status < 300
    except ValueError:
        pass
    return False


def _header(title):
    yield title
    yield '=' * len(title)
    yield ''


def openapihttpdomain(spec, **options):
    if 'examples' in options:
        raise ValueError(
            'Rendering examples is not supported for OpenAPI v2.x specs.')

    if 'request' in options:
        raise ValueError(
            'The :request: option is not supported for OpenAPI v2.x specs.')

    generators = []

    # OpenAPI spec may contain JSON references, common properties, etc.
    # Trying to render the spec "As Is" will require to put multiple
    # if-s around the code. In order to simplify flow, let's make the
    # spec to have only one (expected) schema, i.e. normalize it.
    utils.normalize_spec(spec, **options)

    # Paths list to be processed
    paths = []

    # If 'paths' are passed we've got to ensure they exist within an OpenAPI
    # spec; otherwise raise error and ask user to fix that.
    if 'paths' in options:
        if not set(options['paths']).issubset(spec['paths']):
            raise ValueError(
                'One or more paths are not defined in the spec: %s.' % (
                    ', '.join(set(options['paths']) - set(spec['paths'])),
                )
            )
        paths = options['paths']

    # Check against regular expressions to be included
    if 'include' in options:
        for i in options['include']:
            ir = re.compile(i)
            for path in spec['paths']:
                if ir.match(path):
                    paths.append(path)

    # If no include nor paths option, then take full path
    if 'include' not in options and 'paths' not in options:
        paths = spec['paths']

    # Remove paths matching regexp
    if 'exclude' in options:
        _paths = []
        for e in options['exclude']:
            er = re.compile(e)
            for path in paths:
                if not er.match(path):
                    _paths.append(path)
        paths = _paths

    if 'group' in options:
        groups = collections.OrderedDict(
            [(x['name'], []) for x in spec.get('tags', {})]
            )

        for endpoint in paths:
            for method, properties in spec['paths'][endpoint].items():
                if options.get('methods') and method not in options.get('methods'):
                    continue
                key = properties.get('tags', [''])[0]
                groups.setdefault(key, []).append(_httpresource(
                    endpoint,
                    method,
                    properties,
                    utils.get_text_converter(options),
                    ))

        for key in groups.keys():
            if key:
                generators.append(_header(key))
            else:
                generators.append(_header('default'))

            generators.extend(groups[key])
    else:
        for endpoint in paths:
            for method, properties in spec['paths'][endpoint].items():
                if options.get('methods') and method not in options.get('methods'):
                    continue
                generators.append(_httpresource(
                    endpoint,
                    method,
                    properties,
                    utils.get_text_converter(options),
                    ))

    return iter(itertools.chain(*generators))



"""Here lies still breathing and only renderer implementation."""

from docutils.parsers.rst import directives

from . import abc
from .. import openapi20, openapi30, utils


class HttpdomainOldRenderer(abc.RestructuredTextRenderer):

    option_spec = {
        # A list of endpoints to be rendered. Endpoints must be whitespace
        # delimited.
        "paths": lambda s: s.split(),
        # Regular expression patterns to includes/excludes endpoints to/from
        # rendering. Similar to paths, the patterns must be whitespace
        # delimited.
        "include": lambda s: s.split(),
        "exclude": lambda s: s.split(),
        # Endpoints to be included based on HTTP method names.
        "methods": lambda s: s.split(),
        # Render the request body structure when passed.
        "request": directives.flag,
        # Render request/response examples when passed.
        "examples": directives.flag,  # render examples when passed
        # Render request/response examples in one block or inline with the response codes.
        "group_examples": directives.flag,  # render examples in one block
        # Include links to entity description
        "entities": directives.flag,
        # Group endpoints by tags when passed. By default, no grouping is
        # applied and endpoints are rendered in the order they met in spec.
        "group": directives.flag,
        # Markup format to render OpenAPI descriptions.
        "format": str,
    }

    def __init__(self, state, options):
        self._state = state
        self._options = options

    def render_restructuredtext_markup(self, spec):
        # OpenAPI spec may contain JSON references, common properties, etc.
        # Trying to render the spec "As Is" will require to put multiple if-s
        # around the code. In order to simplify rendering flow, let's make it
        # have only one (expected) schema, i.e. normalize it.
        utils.normalize_spec(spec, **self._options)

        # We support both OpenAPI 2.0 (f.k.a. Swagger) and OpenAPI 3.0.0, so
        # determine which version we are parsing here.
        spec_version = spec.get("openapi", spec.get("swagger", "2.0"))
        if spec_version.startswith("2."):
            openapihttpdomain = openapi20.openapihttpdomain
        elif spec_version.startswith("3."):
            openapihttpdomain = openapi30.openapihttpdomain
        else:
            raise ValueError("Unsupported OpenAPI version (%s)" % spec_version)

        yield from openapihttpdomain(spec, **self._options)



"""Some shared goodies."""

import textwrap

import pytest
import yaml

from sphinxcontrib.openapi import renderers


@pytest.fixture(scope="function")
def fakestate():
    return None


@pytest.fixture(scope="function")
def testrenderer(fakestate):
    return renderers.HttpdomainRenderer(fakestate, {})


@pytest.fixture(scope="function")
def oas_fragment():
    def oas_fragment(fragment):
        return yaml.safe_load(textwrap.dedent(fragment))

    return oas_fragment



"""Smoke test examples from OAI/OpenAPI-Specification repo."""

import os
import glob
import itertools

import py
import pytest


@pytest.mark.parametrize('spec', itertools.chain(
    glob.glob(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'OpenAPI-Specification',
            'examples',
            'v2.0',
            'json',
            '*.json')),

    glob.glob(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'OpenAPI-Specification',
            'examples',
            'v2.0',
            'yaml',
            '*.yaml')),
))
def test_openapi2_success(tmpdir, run_sphinx, spec):
    py.path.local(spec).copy(tmpdir.join('src', 'test-spec.yml'))
    run_sphinx('test-spec.yml')


@pytest.mark.parametrize('group_paths', [False, True])
@pytest.mark.parametrize('render_examples', [False, True])
@pytest.mark.parametrize('render_request', [False, True])
@pytest.mark.parametrize('spec', glob.glob(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        'OpenAPI-Specification',
        'examples',
        'v3.0',
        '*.yaml')
))
def test_openapi3_success(tmpdir, run_sphinx, spec, render_examples,
                          render_request, group_paths):
    py.path.local(spec).copy(tmpdir.join('src', 'test-spec.yml'))
    run_sphinx('test-spec.yml', options={
        'examples': render_examples,
        'request': render_request,
        'group': group_paths,
    })



""".convert_parameter() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_parameter(oas_fragment):
    def _wrapper(parameter):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"]["parameters"] = [parameter]

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"]["parameters"][0]

    return _wrapper


def test_in_header_complete(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            description: token to be passed as a header
            in: header
            items:
              format: int64
              type: integer
            name: token
            required: true
            type: array
            """
        )
    )
    assert converted == oas_fragment(
        """
        description: token to be passed as a header
        in: header
        name: token
        required: true
        schema:
          items:
            format: int64
            type: integer
          type: array
        """
    )


def test_in_path_complete(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            description: username to fetch
            in: path
            name: username
            required: true
            type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        description: username to fetch
        in: path
        name: username
        required: true
        schema:
          type: string
        """
    )


def test_in_query_complete(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            description: ID of the object to fetch
            in: query
            items:
              type: string
            name: id
            required: false
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        description: ID of the object to fetch
        in: query
        name: id
        required: false
        schema:
          items:
            type: string
          type: array
        """
    )


def test_in_header_minimal(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            in: header
            name: token
            type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: header
        name: token
        schema:
          type: string
        """
    )


def test_in_path_minimal(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            in: path
            name: username
            required: true
            type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: path
        name: username
        required: true
        schema:
          type: string
        """
    )


def test_in_query_minimal(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            in: query
            name: id
            type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: query
        name: id
        schema:
          type: string
        """
    )


def test_collectionFormat_is_csv_path(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: csv
            in: path
            items:
              type: string
            name: username
            required: true
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: path
        name: username
        required: true
        schema:
          items:
            type: string
          type: array
        style: simple
        """
    )


def test_collectionFormat_is_csv_header(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: csv
            in: header
            items:
              type: string
            name: username
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: header
        name: username
        schema:
          items:
            type: string
          type: array
        style: simple
        """
    )


def test_collectionFormat_is_csv(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: csv
            in: query
            items:
              type: string
            name: id
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        explode: false
        in: query
        name: id
        schema:
          items:
            type: string
          type: array
        style: form
        """
    )


def test_collectionFormat_is_multi(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: multi
            in: query
            items:
              type: string
            name: id
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        explode: true
        in: query
        name: id
        schema:
          items:
            type: string
          type: array
        style: form
        """
    )


def test_collectionFormat_is_ssv(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: ssv
            in: query
            items:
              type: string
            name: id
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: query
        name: id
        schema:
          items:
            type: string
          type: array
        style: spaceDelimited
        """
    )


def test_collectionFormat_is_pipes(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: pipes
            in: query
            items:
              type: string
            name: id
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: query
        name: id
        schema:
          items:
            type: string
          type: array
        style: pipeDelimited
        """
    )


def test_collectionFormat_is_tsv(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            collectionFormat: tsv
            in: query
            items:
              type: string
            name: id
            type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: query
        name: id
        schema:
          items:
            type: string
          type: array
        """
    )


def test_in_header_vendor_extensions(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            in: header
            name: token
            type: string
            x-vendor-ext: vendor-ext
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: header
        name: token
        schema:
          type: string
        x-vendor-ext: vendor-ext
        """
    )


def test_in_path_vendor_extensions(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            in: path
            name: username
            required: true
            type: string
            x-vendor-ext: vendor-ext
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: path
        name: username
        required: true
        schema:
          type: string
        x-vendor-ext: vendor-ext
        """
    )


def test_in_query_vendor_extensions(convert_parameter, oas_fragment):
    converted = convert_parameter(
        oas_fragment(
            """
            in: query
            name: id
            type: string
            x-vendor-ext: vendor-ext
            """
        ),
    )
    assert converted == oas_fragment(
        """
        in: query
        name: id
        schema:
          type: string
        x-vendor-ext: vendor-ext
        """
    )



""".convert_request_body() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


_MISSING = object()


@pytest.fixture(scope="function")
def convert_request_body(oas_fragment):
    def _wrapper(operation_fragment):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"].update(operation_fragment)

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"].get("requestBody", _MISSING)

    return _wrapper


def test_minimal(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - application/json
            parameters:
              - in: body
                name: user
                schema:
                  $ref: '#/definitions/User'
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            schema:
              $ref: '#/definitions/User'
        """
    )


def test_complete(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            consumes:
              - application/json
            parameters:
              - description: user to add to the system
                in: body
                name: user
                required: true
                schema:
                  $ref: '#/definitions/User'
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          application/json:
            schema:
              $ref: '#/definitions/User'
        description: user to add to the system
        required: true
        """
    )


def test_no_consumes(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - in: body
                name: user
                schema:
                  $ref: '#/definitions/User'
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          '*/*':
            schema:
              $ref: '#/definitions/User'
        """
    )


def test_header_path_query_are_ignored(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - in: header
                name: token
                type: string
              - in: path
                name: username
                required: true
                type: string
              - in: query
                name: id
                type: string
            """
        ),
    )
    assert converted is _MISSING


def test_body_and_others(convert_request_body, oas_fragment):
    converted = convert_request_body(
        oas_fragment(
            """
            parameters:
              - in: header
                name: token
                type: string
              - in: path
                name: username
                required: true
                type: string
              - in: body
                name: user
                schema:
                  $ref: '#/definitions/User'
              - in: query
                name: id
                type: string
            """
        ),
    )
    assert converted == oas_fragment(
        """
        content:
          '*/*':
            schema:
              $ref: '#/definitions/User'
        """
    )
